"""
========
simulate
========

Simulate data.

"""

import collections
import itertools
import math
import random
import re
import tempfile

import pandas as pd

import plotnine as p9

import scipy

import dms_variants.codonvarianttable
from dms_variants.constants import (AAS_WITHSTOP,
                                    CBPALETTE,
                                    CODONS,
                                    CODON_TO_AA,
                                    NTS,
                                    )


def rand_seq(seqlen, *, exclude=None, nseqs=1):
    """Random nucleotide sequence(s).

    Parameters
    ----------
    seqlen : int
        Length of sequence(s).
    exclude : None or list
        Do not generate any of the sequences in this list.
    nseqs : int
        If >1, return a list of this many sequences.

    Returns
    -------
    str or list
        A random sequence or list of them depending on value of `nseqs`.
        If multiple sequences, they are all unique.

    Seed `random.seed` for reproducible output.

    Example
    -------
    >>> random.seed(1)
    >>> rand_seq(4)
    'CAGA'

    >>> random.seed(1)
    >>> rand_seq(5, nseqs=3)
    ['CAGAT', 'TTTCA', 'TATTA']

    >>> random.seed(1)
    >>> rand_seq(5, nseqs=3, exclude=['TTTCA'])
    ['CAGAT', 'TATTA', 'TGCAG']

    """
    if exclude:
        exclude = {s.upper() for s in exclude}
        if len(exclude) >= len(NTS)**seqlen:
            raise ValueError('excluding too many sequences, none valid.')
    else:
        exclude = set()
    seqs = []
    while len(seqs) < nseqs:
        seq = ''.join(random.choice(NTS) for _ in range(seqlen))
        if seq not in exclude:
            seqs.append(seq)
            exclude.add(seq)
    if len(seqs) == 1:
        return seqs[0]
    else:
        return seqs


def mutate_seq(seq, nmuts):
    """Mutate a nucleotide sequence.

    Parameters
    ----------
    seq : str
        Sequence to mutate.
    nmuts: int
        Number of mutations to make to sequence.

    Returns
    -------
    str
        Mutagenized sequence, all upper-case.

    Seed `random.seed` for reproducible output.

    Example
    -------
    >>> random.seed(1)
    >>> mutate_seq('ATGCAA', 2)
    'AAGCGA'

    """
    if nmuts > len(seq):
        raise ValueError(f"Can not make {nmuts} mutations to sequence of "
                         f"length {len(seq)}.")
    seq = list(seq.upper())
    sites = random.sample(range(len(seq)), nmuts)
    for r in sites:
        seq[r] = random.choice([nt for nt in NTS if nt != seq[r]])
    return ''.join(seq)


def codon_muts(codonseq, nmuts, nvariants):
    """Get random codon mutations to sequence.

    Parameters
    ----------
    codonseq : str
        A valid codon sequence
    nmuts : int
        Number of mutations to make per variant.
    nvariants : int
        Number of variants to generate. There is no guarantee that the
        variants are unique.

    Returns
    -------
    list
        List of mutations in each of the `nvariants` mutated variants
        in the form 'ATG1TAC GGC3GGG' where the number refers to the
        **codon** position.

    Seed `random.seed` for reproducible output.

    Example
    -------
    >>> codonseq = 'ATGGAAGGGCATGGA'
    >>> random.seed(1)
    >>> codon_muts(codonseq, nmuts=2, nvariants=3)
    ['ATG1CAC GAA2ACT', 'CAT4CTT GGA5GGG', 'GAA2ACG CAT4GAA']

    """
    if not re.fullmatch(f"(?:{'|'.join(CODONS)})+", codonseq):
        raise ValueError(f"`codonseq` not valid codon sequence:\n{codonseq}")
    assert len(codonseq) % 3 == 0
    codons = {i + 1: codonseq[3 * i: 3 * i + 3]
              for i in range(len(codonseq) // 3)}
    if nmuts > len(codons):
        raise ValueError('`nmuts` greater than number of codons in `codonseq`')
    mutlist = []
    for _ in range(nvariants):
        sitemuts = []
        for site, wt in sorted(random.sample(codons.items(), nmuts)):
            mut = random.choice([c for c in CODONS if c != wt])
            sitemuts.append(f"{wt}{site}{mut}")
        mutlist.append(' '.join(sitemuts))
    return mutlist


def simulate_CodonVariantTable(*, geneseq, bclen, library_specs,
                               seed=1, variant_call_support=0.5):
    """Simulate :class:`dms_variants.codonvarianttable.CodonVariantTable`.

    Note
    ----
    Only simulates the variants, not counts for samples. To simulate counts,
    use :func:`simulateSampleCounts`.

    Parameters
    -----------
    geneseq : str
        Sequence of wildtype protein-coding gene.
    bclen : int
        Length of the barcodes; must enable complexity at least
        10-fold greater than max number of variants.
    library_specs : dict
        Specifications for each simulated library. Keys are 'avgmuts'
        and 'nvariants'. Mutations per variant are Poisson distributed.
    seed : int or None
        Random number seed or `None` to set no seed.
    variant_call_support : int > 0
        Support is floor of 1 plus draw from exponential distribution
        with this mean.

    Returns
    -------
    :class:`dms_variants.codonvarianttable.CodonVariantTable`

    """
    if seed is not None:
        scipy.random.seed(seed)
        random.seed(seed)

    if len(library_specs) < 1:
        raise ValueError('empty `library_specs`')

    if variant_call_support <= 0:
        raise ValueError('`variant_call_support` must be int > 0')

    if len(geneseq) % 3 != 0:
        raise ValueError('length of `geneseq` not multiple of 3')
    genelength = len(geneseq) // 3

    barcode_variant_dict = collections.defaultdict(list)
    for lib, specs_dict in sorted(library_specs.items()):

        nvariants = specs_dict['nvariants']
        avgmuts = specs_dict['avgmuts']
        if 10 * nvariants > (len(NTS))**bclen:  # safety factor 10
            raise ValueError('barcode too short for nvariants')
        existing_barcodes = set()

        for _ivariant in range(nvariants):

            barcode = ''.join(random.choices(NTS, k=bclen))
            while barcode in existing_barcodes:
                barcode = ''.join(random.choices(NTS, k=bclen))
            existing_barcodes.add(barcode)

            support = math.floor(1 + random.expovariate(variant_call_support))

            # get mutations
            substitutions = []
            nmuts = scipy.random.poisson(avgmuts)
            for icodon in random.sample(range(1, genelength + 1), nmuts):
                wtcodon = geneseq[3 * (icodon - 1): 3 * icodon]
                mutcodon = random.choice([c for c in CODONS if c != wtcodon])
                for i_nt, (wt_nt, mut_nt) in enumerate(zip(wtcodon, mutcodon)):
                    if wt_nt != mut_nt:
                        igene = 3 * (icodon - 1) + i_nt + 1
                        substitutions.append(f"{wt_nt}{igene}{mut_nt}")
            substitutions = ' '.join(substitutions)

            barcode_variant_dict['barcode'].append(barcode)
            barcode_variant_dict['substitutions'].append(substitutions)
            barcode_variant_dict['library'].append(lib)
            barcode_variant_dict['variant_call_support'].append(support)

    barcode_variants = pd.DataFrame(barcode_variant_dict)

    with tempfile.NamedTemporaryFile(mode='w') as f:
        barcode_variants.to_csv(f, index=False)
        f.flush()
        cvt = dms_variants.codonvarianttable.CodonVariantTable(
                        barcode_variant_file=f.name,
                        geneseq=geneseq)

    return cvt


def simulateSampleCounts(*,
                         variants,
                         phenotype_func,
                         variant_error_rate,
                         pre_sample,
                         post_samples,
                         secondary_target_phenotypes=None,
                         pre_sample_name='pre-selection',
                         seed=1):
    """Simulate pre- and post-selection variant counts.

    Note
    ----
    Add to :class:`dms_variants.codonvarianttable.CodonVariantTable` using
    :meth:`dms_variants.codonvarianttable.CodonVariantTable.addSampleCounts`

    Parameters
    ----------
    variants : class:`dms_variants.codonvarianttable.CodonVariantTable`
        Holds variants used in simulation.
    phenotype_func : function
        Takes a space-delimited string of amino-acid substitutions and returns
        a number giving enrichment of variant relative to wildtype. For
        instance, :meth:`SigmoidPhenotypeSimulator.observedEnrichment`.
        If there are multiple targets in `variants`, this function is only
        applied to the primary target. See `secondary_target_phenotypes` for
        the secondary targets.
    variant_error_rate : float
        Rate at which variants mis-called. Provide the probability that a
        variant has a spuriously called (or missing) codon mutation; each
        variant then has a random codon mutation added or removed with this
        probability before being passed to `phenotype_func`.
    pre_sample : pandas.DataFrame or dict
        Counts of each variant pre-selection. To specify counts, provide
        data frame with columns "library", "barcode", and "count". To
        simulate, provide dict with keys "total_count" and "uniformity",
        and for each library, we simulate pre-selection counts as a draw
        of "total_count" counts from a multinomial parameterized by pre-
        selection frequencies drawn from a Dirichlet distribution with
        concentration parameter of "uniformity" (5 is reasonable value).
    post_samples : dict
        Keyed by name of each sample with value another dict keyed by
        'total_count', 'noise', and 'bottleneck'. Counts drawn from
        multinomial parameterized by pre-selection frerquencies after
        passing through bottleneck of indicated size, and adding noise
        by mutiplying phenotype by a random variable with mean 1 and
        standard deviation specified by 'noise' (0 is no noise).
    secondary_target_phenotypes : None or dict
        If there are secondary targets in `variants`, this should be keyed
        by each secondary target name and give the associated phenotypes
        as values.
    pre_sample_name : str
        Name used for the pre-selection sample.
    seed : None or int
        If not `None`, random number seed.

    Returns
    -------
    pandas.DataFrame
        Data frame with the following columns:
            - "library"
            - "barcode"
            - "sample"
            - "count"

    """
    if seed is not None:
        scipy.random.seed(seed)
        random.seed(seed)

    if pre_sample_name in post_samples:
        raise ValueError('`pre_sample_name` is in `post_samples`')

    # -----------------------------------------
    # internal function
    def _add_variant_errors(codon_substitutions):
        """Add errors to variant according to `variant_error_rate`."""
        if random.random() < variant_error_rate:
            muts = codon_substitutions.split()
            if len(muts) == 0 or random.random() < 0.5:
                # add mutation
                mutatedsites = {int(re.match(f"^({'|'.join(CODONS)})"
                                             r'(?P<r>\d+)'
                                             f"({'|'.join(CODONS)})$",
                                             mut).group('r'))
                                for mut in muts}
                unmutatedsites = [r for r in variants.sites
                                  if r not in mutatedsites]
                if not unmutatedsites:
                    raise RuntimeError("variant already has all mutations")
                errorsite = random.choice(unmutatedsites)
                wtcodon = variants.codons[errorsite]
                mutcodon = random.choice([c for c in CODONS if c != wtcodon])
                muts.append(f'{wtcodon}{errorsite}{mutcodon}')
                return ' '.join(muts)
            else:
                # remove mutation
                _ = muts.pop(random.randint(0, len(muts) - 1))
                return ' '.join(muts)
        else:
            return codon_substitutions
    # -----------------------------------------

    if variants.primary_target is None:
        primary_df = variants.barcode_variant_df.assign(target=None)
        secondary_df = None
    else:
        primary_df = (variants.barcode_variant_df
                      .query('target == @variants.primary_target'))
        secondary_df = (variants.barcode_variant_df
                        .query('target != @variants.primary_target'))

    primary_df = (
        primary_df
        [['target', 'library', 'barcode', 'codon_substitutions']]
        .assign(
            codon_substitutions=(lambda x: x['codon_substitutions']
                                 .map(_add_variant_errors)),
            aa_substitutions=(lambda x: x['codon_substitutions']
                              .map(dms_variants.codonvarianttable
                                   .CodonVariantTable.codonToAAMuts)),
            phenotype=lambda x: x['aa_substitutions'].map(phenotype_func)
            )
        [['library', 'barcode', 'phenotype']]
        )

    if secondary_df is not None:
        if secondary_target_phenotypes is None:
            secondary_target_phenotypes = {}
        if not set(secondary_target_phenotypes).issuperset(
                    secondary_df['target']):
            raise ValueError('`secondary_target_phenotypes` lacks some '
                             'secondary targets')
        secondary_df = (
                secondary_df
                .assign(phenotype=lambda x: x['target'].map(
                                        secondary_target_phenotypes))
                [['library', 'barcode', 'phenotype']]
                )
        barcode_variant_df = primary_df.append(secondary_df)
    else:
        barcode_variant_df = primary_df

    libraries = variants.libraries

    if isinstance(pre_sample, pd.DataFrame):
        # pre-sample counts specified
        req_cols = ['library', 'barcode', 'count']
        if not set(req_cols).issubset(set(pre_sample.columns)):
            raise ValueError(f"pre_sample lacks cols {req_cols}:"
                             f"\n{pre_sample}")
        cols = ['library', 'barcode']
        if any(pre_sample[cols].sort_values(cols).reset_index() !=
               barcode_variant_df[cols].sort_values(cols).reset_index()):
            raise ValueError("pre_sample DataFrame lacks required "
                             "library and barcode columns")
        barcode_variant_df = (
                barcode_variant_df
                .merge(pre_sample[req_cols], on=['library', 'barcode'])
                .rename(columns={'count': pre_sample_name})
                )
        # "true" pre-selection freqs are just input counts
        nperlib = (barcode_variant_df
                   .groupby('library')
                   .rename('total_count')
                   .reset_index()
                   )
        barcode_variant_df = (
                barcode_variant_df
                .merge(nperlib, on='library')
                .assign(pre_freqs=lambda x: x[pre_sample_name] / x.total_count)
                .drop(columns='total_count')
                )

    elif isinstance(pre_sample, dict):
        pre_req_keys = {'uniformity', 'total_count'}
        if set(pre_sample.keys()) != pre_req_keys:
            raise ValueError(f"pre_sample lacks required keys {pre_req_keys}")

        pre_df_list = []
        for lib in libraries:  # noqa: B007
            df = (
                barcode_variant_df
                .query('library == @lib')
                .assign(
                    pre_freq=lambda x: scipy.random.dirichlet(
                             pre_sample['uniformity'] *
                             scipy.ones(len(x))),
                    count=lambda x: scipy.random.multinomial(
                            pre_sample['total_count'], x.pre_freq),
                    sample=pre_sample_name
                    )
                )
            pre_df_list.append(df)
        barcode_variant_df = pd.concat(pre_df_list)

    else:
        raise ValueError("pre_sample not DataFrame / dict: "
                         f"{pre_sample}")

    cols = ['library', 'barcode', 'sample', 'count',
            'pre_freq', 'phenotype']
    assert set(barcode_variant_df.columns) == set(cols), (
            f"cols = {set(cols)}\nbarcode_variant_df.columns = "
            f"{set(barcode_variant_df.columns)}")
    for col in cols:
        if col in post_samples:
            raise ValueError(f"post_samples can't have key {col}; "
                             "choose another sample name")

    df_list = [barcode_variant_df[cols[: 4]]]

    def _bottleneck_freqs(pre_freq, bottleneck):
        if bottleneck is None:
            return pre_freq
        else:
            return scipy.random.multinomial(bottleneck, pre_freq) / bottleneck

    post_req_keys = {'bottleneck', 'noise', 'total_count'}
    for lib, (sample, sample_dict) in itertools.product(  # noqa: B007
            libraries, sorted(post_samples.items())):

        if set(sample_dict.keys()) != post_req_keys:
            raise ValueError(f"post_samples {sample} lacks {post_req_keys}")

        lib_df = (
            barcode_variant_df.query('library == @lib')
            .assign(
                sample=sample,
                # simulated pre-selection freqs after bottleneck
                bottleneck_freq=(lambda x:
                                 _bottleneck_freqs(x.pre_freq,
                                                   sample_dict['bottleneck'])),
                # post-selection freqs with noise
                noise=lambda x: scipy.clip(
                                    [random.gauss(1, sample_dict['noise'])
                                     for _ in range(len(x))],
                                    0, None),
                post_freq_nonorm=lambda x: (x.bottleneck_freq *
                                            x.phenotype * x.noise),
                post_freq=lambda x: (x.post_freq_nonorm /
                                     x.post_freq_nonorm.sum()),
                # post-selection counts simulated from frequencies
                count=(lambda x:
                       scipy.random.multinomial(sample_dict['total_count'],
                                                x.post_freq))
                )
            .rename(columns={'post_counts': sample})
            [['library', 'barcode', 'sample', 'count']]
            )

        df_list.append(lib_df)

    return pd.concat(df_list)


class SigmoidPhenotypeSimulator:
    r"""Simulate phenotypes under sigmoid global epistasis model.

    Note
    ----
    Mutational effects on latent phenotype are simulated to follow
    compound normal distribution; latent phenotype maps to observed
    enrichment via sigmoid, and the observed phenotype is the
    :math:`\log_2` of the enrichment. This distinction between latent and
    observed phenotype parallels the "global epistasis" models of
    `Otwinoski et al <https://doi.org/10.1073/pnas.1804015115>`_ and
    `Sailer and Harms <http://www.genetics.org/content/205/3/1079>`_.

    Specifically, let :math:`p_{\rm{latent}}` be the latent phenotype
    of a variant, let :math:`p_{\rm{observed}}` be the observed phenotype,
    and let :math:`E_{\rm{observed}}` be the observed enrichment. Also
    let :math:`p_{\rm{latent, wt}}` be the latent phenotype of the
    wildtype. Then:

    .. math::

       E_{\rm{observed}}
       &=& \frac{\left(1 + \exp\left(-p_{\rm{latent, wt}}\right)\right)
                 \left(1 - E_{\rm{min}}\right)}
                {1 + \exp\left(-p_{\rm{latent}}\right)} +
                E_{\rm{min}} \\
       p_{\rm{observed}} &=& \log_2 E_{\rm{observed}}

    where :math:`0 \le E_{\rm{min}} < 1` is the minimum possible observed
    enrichment. Typically :math:`E_{\rm{min}} > 0` in real experiments
    as pseudocounts are used so estimates of enrichments can't be zero.

    Parameters
    ----------
    geneseq : str
        Codon sequence of wild-type gene.
    seed : int or None
        Random number seed.
    wt_latent : float
        Latent phenotype of wildtype.
    norm_weights : list or tuple of tuples
        Specify compound normal distribution of mutational effects on
        latent phenotype as `(weight, mean, sd)` for each Gaussian.
    stop_effect : float
        Effect of stop codon at any position.
    min_observed_enrichment : float
        Minimum possible observed enrichment.

    Attributes
    ----------
    wt_latent : float
        Wildtype latent phenotype.
    muteffects : dict
        Effect on latent phenotype of each amino-acid mutation.

    """

    def __init__(self, geneseq, *, seed=1, wt_latent=5,
                 norm_weights=((0.4, -0.7, 1.5), (0.6, -7, 3.5)),
                 stop_effect=-15, min_observed_enrichment=0.001):
        """See main class docstring for how to initialize."""
        self.wt_latent = wt_latent

        if not (0 <= min_observed_enrichment < 1):
            raise ValueError('not 0 <= `min_observed_enrichment` < 1')
        self.min_observed_enrichment = min_observed_enrichment

        # simulate muteffects from compound normal distribution
        self.muteffects = {}
        if seed is not None:
            random.seed(seed)
        weights, means, sds = zip(*norm_weights)
        cumweights = scipy.cumsum(weights)
        for icodon in range(len(geneseq) // 3):
            wt_aa = CODON_TO_AA[geneseq[3 * icodon: 3 * icodon + 3]]
            for mut_aa in AAS_WITHSTOP:
                if mut_aa != wt_aa:
                    if mut_aa == '*':
                        muteffect = stop_effect
                    else:
                        # choose Gaussian from compound normal
                        i = scipy.argmin(cumweights < random.random())
                        # draw mutational effect from chosen Gaussian
                        muteffect = random.gauss(means[i], sds[i])
                    self.muteffects[f"{wt_aa}{icodon + 1}{mut_aa}"] = muteffect

    def latentPhenotype(self, subs):
        """Latent phenotype of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.

        Returns
        -------
        float
            Latent phenotype of variant.

        """
        return self.wt_latent + sum(self.muteffects[m] for m in subs.split())

    def observedPhenotype(self, subs):
        """Observed phenotype of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.

        Returns
        -------
        float
            Observed phenotype of variant.

        """
        return self.latentToObserved(self.latentPhenotype(subs), 'phenotype')

    def observedEnrichment(self, subs):
        """Observed enrichment of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.

        Returns
        -------
        float
            Observed enrichment relative to wildtype.

        """
        return self.latentToObserved(self.latentPhenotype(subs), 'enrichment')

    def latentToObserved(self, latent, value):
        r"""Observed enrichment or observed phenotype from latent phenotype.

        Parameters
        ----------
        latent : float or numpy.ndarray
            The latent phenotype.
        value : {'enrichment', 'phenotype'}
            Get the observed enrichment or the observed phenotype?

        Returns
        -------
        float or numpy.ndarray
            The observed enrichment or phenotype.

        """
        enrichment = ((1 + scipy.exp(-self.wt_latent)) *
                      (1 - self.min_observed_enrichment) /
                      (1 + scipy.exp(-latent)) +
                      self.min_observed_enrichment)
        if value == 'enrichment':
            return enrichment
        elif value == 'phenotype':
            return scipy.log(enrichment) / scipy.log(2)
        else:
            raise ValueError(f"invalid `value` of {value}")

    def plotLatentVsObserved(self, value, *, latent_min=-15,
                             latent_max=10, npoints=200,
                             wt_vline=True):
        """Plot observed enrichment/phenotype as function of latent phenotype.

        Parameters
        ----------
        value : {'enrichment', 'phenotype'}
            Do we plot observed enrichment or observed phenotype?
        latent_min : float
            Smallest value of latent phenotype on plot.
        latent_max : float
            Largest value of latent phenotype on plot.
        npoints : int
            Plot a line fit to this many points.
        wt_vline : bool
            Draw a vertical line at the wildtype latent phenotype.

        Returns
        -------
        plotnine.ggplot.ggplot
            Plot of observed enrichment or phenotype as function of latent
            phenotype.

        """
        latent = scipy.linspace(latent_min, latent_max, npoints)
        observed = self.latentToObserved(latent, value)

        p = (p9.ggplot(pd.DataFrame({'latent': latent, 'observed': observed}),
                       p9.aes('latent', 'observed')
                       ) +
             p9.geom_line() +
             p9.theme(figure_size=(3.5, 2.5)) +
             p9.xlab('latent phenotype') +
             p9.ylab(f"observed {value}")
             )

        if wt_vline:
            p = p + p9.geom_vline(xintercept=self.wt_latent,
                                  color=CBPALETTE[1],
                                  linetype='dashed')

        return p

    def plotMutsHistogram(self, value, *,
                          mutant_order=1, bins=30, wt_vline=True):
        """Plot distribution of phenotype for all mutants of a given order.

        Parameters
        ----------
        value : {'latentPhenotype', 'observedPhenotype', 'observedEnrichment'}
            What value to plot.
        mutant_order : int
            Plot mutations of this order. Currently only works for 1
            (single mutants).
        bins : int
            Number of bins in histogram.
        wt_vline : bool
            Draw a vertical line at the wildtype value.

        Returns
        -------
        plotnine.ggplot.ggplot
            Histogram of phenotype for all mutants.

        """
        if mutant_order != 1:
            raise ValueError('only implemented for `mutant_order` of 1')

        if value not in {'latentPhenotype', 'observedPhenotype',
                         'observedEnrichment'}:
            raise ValueError(f"invalid `value` of {value}")
        func = getattr(self, value)

        xlist = [func(m) for m in self.muteffects.keys()]

        p = (p9.ggplot(pd.DataFrame({value: xlist}),
                       p9.aes(value)) +
             p9.geom_histogram(bins=bins) +
             p9.theme(figure_size=(3.5, 2.5)) +
             p9.ylab(f"number of {mutant_order}-mutants")
             )

        if wt_vline:
            p = p + p9.geom_vline(
                        xintercept=func(''),
                        color=CBPALETTE[1],
                        linetype='dashed')

        return p


class MultiLatentSigmoidPhenotypeSimulator:
    r"""Simulate phenotype that derives from several latent phenotypes.

    Note
    -----
    The observed phenotype is the sum of several sigmoid-transformed latent
    phenotypes. Specifically, let there be :math:`k = 1, \ldots, K` latent
    phenotypes, each of which is transformed into an observed phenotype
    :math:`p_{\rm{observed}}^k as described in the docs for
    :class:`SigmoidPhenotypeSimulator`. The overall observed phenotype is:

    .. math::

       p_{\rm{observed}} = \sum_k p_{\rm{observed}}^k

    and the observed enrichment is:

    .. math::

       E_{\rm{observed}} = 2^{p_{\rm{observed}}}.

    Parameters
    -----------
    sigmoid_phenotype_simulators : list
        A list of :class:`SigmoidPhenotypeSimulator` objects, each of which
        transforms one of the latent phenotypes to a component of the
        observed phenotype.

    Attributes
    ----------
    n_latent_phenotypes : int
        The number of latent phenotypes.

    """

    def __init__(self, sigmoid_phenotype_simulators):
        """See main class docstring."""
        self.n_latent_phenotypes = len(sigmoid_phenotype_simulators)
        if self.n_latent_phenotypes < 1:
            raise ValueError('`sigmoid_phenotype_simulators` cannot be empty')
        if not all(isinstance(m, SigmoidPhenotypeSimulator) for m in
                   sigmoid_phenotype_simulators):
            raise ValueError('entries in `sigmoid_phenotype_simulators` not '
                             'all of SigmoidPhenotypeSimulator type')
        self._sigmoid_phenotype_simulators = sigmoid_phenotype_simulators
        self._all_subs = list(self._sigmoid_phenotype_simulators[0]
                              .muteffects.keys())
        for m in self._sigmoid_phenotype_simulators[1:]:
            if set(self._all_subs) != set(m.muteffects.keys()):
                raise ValueError('entries in `sigmoid_phenotype_simulators` '
                                 'do not all represent same gene sequence')

    def observedEnrichment(self, subs):
        """Observed enrichment of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.

        Returns
        -------
        float
            Observed enrichment relative to wildtype.

        """
        return 2**self.observedPhenotype(subs)

    def observedPhenotype(self, subs):
        """Observed phenotype of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.

        Returns
        -------
        float
            Observed phenotype of a variant.

        """
        return sum(m.observedPhenotype(subs) for m in
                   self._sigmoid_phenotype_simulators)

    def latentPhenotype(self, subs, k):
        """Latent phenotype :math:`k` of a variant.

        Parameters
        ----------
        subs : str
            Space-delimited list of amino-acid substitutions.
        k : int
            Latent phenotype to get (1 <= `k` <=
            :attr:`MultiLatentSigmoidPhenotypeSimulator.n_latent_phenotypes`).

        Returns
        -------
        float
            Latent phenotype `k`.

        """
        if isinstance(k, int) and 1 <= k <= self.n_latent_phenotypes:
            return (self._sigmoid_phenotype_simulators[k - 1]
                    .observedPhenotype(subs))
        else:
            raise ValueError(f"invalid `k` of {k}")

    def plotMutsHistogram(self,
                          value, *,
                          k=None,
                          mutant_order=1,
                          bins=30,
                          wt_vline=True,
                          ):
        """Plot distribution of phenotype for all mutants of a given order.

        Parameters
        ----------
        value : {'latentPhenotype', 'observedPhenotype', 'observedEnrichment'}
            What value to plot.
        k : int or None
            If value is `latentPhenotype, which phenotype (1 <= `k` <=
            :attr:`MultiLatentSigmoidPhenotypeSimulator.n_latent_phenotypes`)
            to plot.
        mutant_order : int
            Plot mutations of this order. Currently only works for 1
            (single mutants).
        bins : int
            Number of bins in histogram.
        wt_vline : bool
            Draw a vertical line at the wildtype value.

        Returns
        -------
        plotnine.ggplot.ggplot
            Histogram of phenotype for all mutants.

        """
        if mutant_order != 1:
            raise ValueError('only implemented for `mutant_order` of 1')

        if value == 'latentPhenotype':
            if isinstance(k, int) and 1 <= k <= self.n_latent_phenotypes:
                kwargs = {'k': k}
                xlabel = f"latentPhenotype {k}"
            else:
                raise ValueError(f"invalid `k` of {k}")
        else:
            kwargs = {}
            xlabel = value

        if value not in {'latentPhenotype', 'observedPhenotype',
                         'observedEnrichment'}:
            raise ValueError(f"invalid `value` of {value}")
        func = getattr(self, value)

        xlist = [func(m, **kwargs) for m in self._all_subs]

        p = (p9.ggplot(pd.DataFrame({value: xlist}),
                       p9.aes(value)) +
             p9.geom_histogram(bins=bins) +
             p9.theme(figure_size=(3.5, 2.5)) +
             p9.ylab(f"number of {mutant_order}-mutants") +
             p9.xlab(xlabel)
             )

        if wt_vline:
            p = p + p9.geom_vline(
                        xintercept=func('', **kwargs),
                        color=CBPALETTE[1],
                        linetype='dashed')

        return p


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
================================
dms_variants
================================

Python package for analyzing deep mutational scanning of barcoded
variants of genes.

"""

__author__ = '`the Bloom lab <https://research.fhcrc.org/bloom/en.html>`_'
__email__ = 'jbloom@fredhutch.org'
__version__ = '0.7.1'
__url__ = 'https://github.com/jbloomlab/dms_variants'



"""
============
constants
============

Defines constants used by package.

"""


import Bio.Seq


CBPALETTE = ('#999999', '#E69F00', '#56B4E9', '#009E73',
             '#F0E442', '#0072B2', '#D55E00', '#CC79A7')
"""tuple: Color-blind safe palette.

From http://bconnelly.net/2013/10/creating-colorblind-friendly-figures/
"""

AAS_NOSTOP = ('A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L',
              'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y')
"""tuple: Amino-acid one-letter codes alphabetized, doesn't include stop."""

AAS_WITHSTOP = tuple(list(AAS_NOSTOP) + ['*'])
"""tuple: Amino-acid one-letter codes alphabetized plus stop as ``*``."""

NTS = ('A', 'C', 'G', 'T')
"""tuple: DNA nucleotide one-letter codes."""

NTS_AMBIGUOUS = ('A', 'B', 'C', 'D', 'G', 'H', 'K', 'M', 'N',
                 'R', 'S', 'T', 'V', 'W', 'Y')
"""tuple: DNA nucleotide one-letter codes including ambiguous ones."""

NT_COMPLEMENT = {_nt: str(Bio.Seq.Seq(_nt).reverse_complement()) for
                 _nt in NTS_AMBIGUOUS}
"""dict: Maps each nucleotide to its complement, including ambiguous ones."""

NT_TO_REGEXP = dict(map(lambda tup: ((tup[0], tup[1]) if len(tup[1]) == 1 else
                                     (tup[0], '[' + ''.join(tup[1]) + ']')),
                        Bio.Seq.IUPAC.IUPACData.ambiguous_dna_values.items()
                        ))
"""dict: Maps nucleotide code to regular expression expansion."""

CODONS = tuple(f"{_n1}{_n2}{_n3}" for _n1 in NTS for _n2 in NTS for _n3 in NTS)
"""tuple: DNA codons, alphabetized."""

CODON_TO_AA = {_c: str(Bio.Seq.Seq(_c).translate()) for _c in CODONS}
"""dict: Maps codons to amino acids."""

AA_TO_CODONS = {_aa: [_c for _c in CODONS if CODON_TO_AA[_c] == _aa]
                for _aa in AAS_WITHSTOP}
"""dict: Reverse translate amino acid to list of encoding codons."""

CODONS_NOSTOP = tuple(_c for _c in CODONS if CODON_TO_AA[_c] != '*')
"""tuple: DNA codons except for stop codons, alphabetized."""



"""
===========
fastq
===========

Tools for processing FASTQ files.

"""


import collections
import gzip
import itertools
import os
import tempfile  # noqa: F401

import numpy


def qual_str_to_array(q_str, *, offset=33):
    """Convert quality score string to array of integers.

    Parameters
    ----------
    q_str : str
        Quality score string.
    offset : int
        Offset in ASCII encoding of Q-scores.

    Returns
    -------
    numpy.ndarray
        Array of integer quality scores.

    Example
    -------
    >>> qual_str_to_array('!I:0G')
    array([ 0, 40, 25, 15, 38])

    """
    return numpy.array([ord(q) - offset for q in q_str],
                       dtype='int')


def iterate_fastq_pair(r1filename, r2filename, *,
                       r1trim=None, r2trim=None, qual_format='str'):
    r"""Iterate over paired R1 and R2 FASTQ files.

    Parameters
    ----------
    r1filename : str
        R1 FASTQ file name, can be gzipped (extension ``.gz``).
    r2filename : str
        R2 FASTQ file name, can be gzipped (extension ``.gz``).
    r1trim : int or None
        If not `None`, trim R1 reads and Q scores to be longer than this.
    r2trim : int or None
        If not `None`, trim R2 reads and Q scores to be longer than this.
    qual_format : {'str', 'array'}
        Return the quality scores as string of ASCII codes or array of numbers?

    Yields
    ------
    namedtuple
        The entries in the tuple are (in order):
         - `id` : read id
         - `r1_seq` : R1 read sequence
         - `r2_seq` : R2 read sequence
         - `r1_qs` : R1 Q scores (`qual_format` parameter determines format)
         - `r2_qs` : R2 Q scores (`qual_format` parameter determines format)
         - `fail` : did either read fail chastity filter? (`None` if no info)

    Example
    -------
    >>> f1 = tempfile.NamedTemporaryFile(mode='w')
    >>> _ = f1.write(
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984 1:N:0:CGATGT\n'
    ...         'ATGCAATTG\n'
    ...         '+\n'
    ...         'GGGGGIIII\n'
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985 1:Y:0:CGATGT\n'
    ...         'ACGCTATTC\n'
    ...         '+\n'
    ...         'GHGGGIKII\n'
    ...         )
    >>> f1.flush()
    >>> f2 = tempfile.NamedTemporaryFile(mode='w')
    >>> _ = f2.write(
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984 2:N:0:CGATGT\n'
    ...         'CAGCATA\n'
    ...         '+\n'
    ...         'AGGGGII\n'
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985 2:Y:0:CGATGT\n'
    ...         'CTGAATA\n'
    ...         '+\n'
    ...         'GHBGGIK\n'
    ...         )
    >>> f2.flush()

    >>> for tup in iterate_fastq_pair(f1.name, f2.name, r1trim=8, r2trim=5,
    ...                               qual_format='array'):
    ...     print(tup)
    ... # doctest: +NORMALIZE_WHITESPACE
    FastqPairEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984',
                   r1_seq='ATGCAATT',
                   r2_seq='CAGCA',
                   r1_qs=array([38, 38, 38, 38, 38, 40, 40, 40]),
                   r2_qs=array([32, 38, 38, 38, 38]),
                   fail=False)
    FastqPairEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985',
                   r1_seq='ACGCTATT',
                   r2_seq='CTGAA',
                   r1_qs=array([38, 39, 38, 38, 38, 40, 42, 40]),
                   r2_qs=array([38, 39, 33, 38, 38]),
                   fail=True)

    >>> for tup in iterate_fastq_pair(f1.name, f2.name, r1trim=8, r2trim=5):
    ...     print(tup)
    ... # doctest: +NORMALIZE_WHITESPACE
    FastqPairEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984',
                   r1_seq='ATGCAATT',
                   r2_seq='CAGCA',
                   r1_qs='GGGGGIII',
                   r2_qs='AGGGG',
                   fail=False)
    FastqPairEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985',
                   r1_seq='ACGCTATT',
                   r2_seq='CTGAA',
                   r1_qs='GHGGGIKI',
                   r2_qs='GHBGG',
                   fail=True)


    >>> f1.close()
    >>> f2.close()

    """
    FastqPairEntry = collections.namedtuple(
                            'FastqPairEntry',
                            'id r1_seq r2_seq r1_qs r2_qs fail')

    r1_iterator = iterate_fastq(r1filename,
                                trim=r1trim,
                                check_pair=1,
                                qual_format=qual_format)
    r2_iterator = iterate_fastq(r2filename,
                                trim=r2trim,
                                check_pair=2,
                                qual_format=qual_format)

    for r1_entry, r2_entry in itertools.zip_longest(r1_iterator, r2_iterator):

        if (r1_entry is None) or (r2_entry is None):
            raise IOError(f"{r1filename} and {r2filename} have unequal "
                          'number of entries')

        if r1_entry[0] != r2_entry[0]:
            raise IOError(f"{r1filename} and {r2filename} specify different "
                          f"read IDs:\n{r1_entry[0]}\n{r2_entry[0]}")

        yield FastqPairEntry(id=r1_entry[0],
                             r1_seq=r1_entry[1],
                             r2_seq=r2_entry[1],
                             r1_qs=r1_entry[2],
                             r2_qs=r2_entry[2],
                             fail=(r1_entry[3] or r2_entry[3]),
                             )


def iterate_fastq(filename, *, trim=None, check_pair=None, qual_format='str'):
    r"""Iterate over a FASTQ file.

    Parameters
    ----------
    filename : str
        FASTQ file name, can be gzipped (extension ``.gz``).
    trim : int or None
        If not `None`, trim reads and Q scores to be longer than this.
    check_pair : {1, 2, None}
        If not `None`, check reads are read 1 or read 2 if this info given.
        Assumes Casava 1.8 or SRA header format.
    qual_format : {'str', 'array'}
        Return the quality scores as string of ASCII codes or array of numbers?

    Yields
    ------
    namedtuple
        The entries in the tuple are (in order):
         - `id` : read id
         - `seq` : read sequence
         - `qs` : Q scores (`qual_format` parameter determines format)
         - `fail` : did read fail chastity filter? (`None` if no filter info)

    Example
    -------
    >>> f = tempfile.NamedTemporaryFile(mode='w')
    >>> _ = f.write(
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984 1:N:0:CGATGT\n'
    ...         'ATGCAATTG\n'
    ...         '+\n'
    ...         'GGGGGIIII\n'
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985 1:Y:0:CGATGT\n'
    ...         'ACGCTATTC\n'
    ...         '+\n'
    ...         'GHGGGIKII\n'
    ...         '@DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985 2:Y:0:CGATGT\n'
    ...         'ACGCTATTC\n'
    ...         '+\n'
    ...         'GHGGGIKII\n'
    ...         )
    >>> f.flush()

    >>> try:
    ...     for tup in iterate_fastq(f.name, trim=5, check_pair=1):
    ...         print(tup)
    ... except ValueError as e:
    ...    print(f"ValueError: {e}")
    ... # doctest: +NORMALIZE_WHITESPACE
    FastqEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984',
               seq='ATGCA',
               qs='GGGGG',
               fail=False)
    FastqEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985',
               seq='ACGCT',
               qs='GHGGG',
               fail=True)
    ValueError: header not for R1:
    @DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985 2:Y:0:CGATGT

    >>> for tup in iterate_fastq(f.name, qual_format='array'):
    ...    print(tup)
    ... # doctest: +NORMALIZE_WHITESPACE
    FastqEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1984',
               seq='ATGCAATTG',
               qs=array([38, 38, 38, 38, 38, 40, 40, 40, 40]),
               fail=False)
    FastqEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985',
               seq='ACGCTATTC',
               qs=array([38, 39, 38, 38, 38, 40, 42, 40, 40]),
               fail=True)
    FastqEntry(id='DH1DQQN1:933:HMLH5BCXY:1:1101:2165:1985',
               seq='ACGCTATTC',
               qs=array([38, 39, 38, 38, 38, 40, 42, 40, 40]),
               fail=True)

    >>> f.close()

    """
    FastqEntry = collections.namedtuple('FastqEntry',
                                        'id seq qs fail')

    if check_pair is not None:
        check_pair = str(check_pair)
        if check_pair not in {'1', '2'}:
            raise ValueError(f"invalid `check_pair` of {check_pair}")

    if not os.path.isfile(filename):
        raise IOError(f"no FASTQ file {filename}")

    if qual_format == 'array':
        qual_to_array = True
    elif qual_format == 'str':
        qual_to_array = False
    else:
        raise ValueError(f"invalid value for `qual_format`: {qual_format}")

    if os.path.splitext(filename)[1].lower() == '.gz':
        openfunc = gzip.open
    else:
        openfunc = open

    with openfunc(filename, mode='rt') as f:
        head = f.readline()
        while head:
            if head[0] != '@':
                raise IOError(f"id starts with {head[0]}, not @:\n{head}")
            else:
                head = head.rstrip()
                headspl = head[1:].split()
                read_id = headspl[0]
            seq = f.readline().rstrip()
            plusline = f.readline().rstrip()
            qs = f.readline().rstrip()
            if (not seq) or (len(seq) != len(qs)) or (plusline != '+'):
                raise IOError(f"invalid entry for {read_id} in {filename}:\n"
                              f"{head}\n{seq}\n{plusline}\n{qs}")

            # trim last two characters (needed for SRA downloads)
            if read_id[-2:] in {'.1', '.2'}:
                if check_pair and (read_id[-1] != check_pair):
                    raise ValueError(f"header not for R{check_pair}:\n{head}")
                read_id = read_id[: -2]

            if check_pair and len(headspl) > 1 and headspl[1][0] != check_pair:
                raise ValueError(f"header not for R{check_pair}:\n{head}")

            # parse chastity filter assuming CASAVA 1.8 header
            try:
                chastity = headspl[1][2]
                if chastity == 'Y':
                    fail = True
                elif chastity == 'N':
                    fail = False
                else:
                    raise ValueError(f"cannot parse chastity filter in {head}")
            except IndexError:
                fail = None  # header does not specify chastity filter

            if trim is not None:
                seq = seq[: trim]
                qs = qs[: trim]

            if qual_to_array:
                qs = qual_str_to_array(qs)

            yield FastqEntry(id=read_id,
                             seq=seq,
                             qs=qs,
                             fail=fail,
                             )

            head = f.readline()


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
=================
ispline
=================

Implements :class:`Isplines`, which are monotonic spline functions that are
defined in terms of :class:`Msplines`. Also implements :class:`Isplines_total`
for the weighted sum of a :class:`Isplines` family.

See `Ramsay (1988)`_ for details about these splines, and also note the
corrections in the `Praat manual`_ to the errors in the I-spline formula
by `Ramsay (1988)`_.

.. _`Ramsay (1988)`: https://www.jstor.org/stable/2245395
.. _`Praat manual`: http://www.fon.hum.uva.nl/praat/manual/spline.html

"""


import numpy


class Isplines_total:
    r"""Evaluate the weighted sum of an I-spline family (see `Ramsay (1988)`_).

    Parameters
    ----------
    order : int
        Sets :attr:`Isplines_total.order`.
    mesh : array-like
        Sets :attr:`Isplines_total.mesh`.
    x : numpy.ndarray
        Sets :attr:`Isplines_total.x`.

    Attributes
    ----------
    order : int
        See :attr:`Isplines.order`.
    mesh : numpy.ndarray
        See :attr:`Isplines.mesh`.
    n : int
        See :attr:`Isplines.n`.
    lower : float
        See :attr:`Isplines.lower`.
    upper : float
        See :attr:`Isplines.upper`.

    Note
    ----
    Evaluates the full interpolating curve from the I-splines. When
    :math:`x` falls within the lower :math:`L` and upper :math:`U`
    bounds of the range covered by the I-splines (:math:`L \le x \le U`),
    then this curve is defined as:

    .. math::

       I_{\rm{total}}\left(x\right)
       =
       w_{\rm{lower}} + \sum_i w_i I_i\left(x\right).

    When :math:`x` is outside the range of the mesh covered by the splines,
    the values are linearly extrapolated from first derivative at the
    bounds. Specifically, if :math:`x < L` then:

    .. math::

       I_{\rm{total}}\left(x\right)
       =
       I_{\rm{total}}\left(L\right) +
       \left(x - L\right)
       \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                  {\partial y}\right\rvert_{y=L},

    and if :math:`x > U` then:

    .. math::

       I_{\rm{total}}\left(x\right)
       =
       I_{\rm{total}}\left(U\right) +
       \left(x - U\right)
       \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                  {\partial y}\right\rvert_{y=U}.

    Note also that:

    .. math::

       I_{\rm{total}}\left(L\right) &=& w_{\rm{lower}}, \\
       I_{\rm{total}}\left(U\right) &=& w_{\rm{lower}} + \sum_i w_i

    Example
    -------
    Short examples to demonstrate and test :class:`Isplines_total`:

    .. plot::
       :context: reset

       >>> import itertools
       >>> import numpy
       >>> import pandas as pd
       >>> import scipy.optimize
       >>> from dms_variants.ispline import Isplines_total

       >>> order = 3
       >>> mesh = [0.0, 0.3, 0.5, 0.6, 1.0]
       >>> x = numpy.array([0, 0.2, 0.3, 0.4, 0.8, 0.99999])
       >>> isplines_total = Isplines_total(order, mesh, x)
       >>> weights = numpy.array([1.2, 2, 1.2, 1.2, 3, 0]) / 6
       >>> numpy.round(isplines_total.Itotal(weights, w_lower=0), 2)
       array([0.  , 0.38, 0.54, 0.66, 1.21, 1.43])

       Now calculate using some points that require linear extrapolation
       outside the mesh and also have a nonzero `w_lower`:

       >>> x2 = numpy.array([-0.5, -0.25, 0, 0.01, 1.0, 1.5])
       >>> isplines_total2 = Isplines_total(order, mesh, x2)
       >>> numpy.round(isplines_total2.Itotal(weights, w_lower=1), 3)
       array([0.   , 0.5  , 1.   , 1.02 , 2.433, 2.433])

       Test :meth:`Isplines_total.dItotal_dx`:

       >>> x_deriv = numpy.array([-0.5, -0.25, 0, 0.01, 0.5, 0.7, 1.0, 1.5])
       >>> for xval in x_deriv:
       ...     xval = numpy.array([xval])
       ...     def func(xval):
       ...         return Isplines_total(order, mesh, xval).Itotal(weights, 0)
       ...     def dfunc(xval):
       ...         return Isplines_total(order, mesh, xval).dItotal_dx(weights)
       ...     err = scipy.optimize.check_grad(func, dfunc, xval)
       ...     if err > 1e-5:
       ...         raise ValueError(f"excess err {err} for {xval}")

       >>> (isplines_total.dItotal_dw_lower() == numpy.ones(x.shape)).all()
       True

       Test :meth:`Isplines_total.dItotal_dweights`:

       >>> isplines_total3 = Isplines_total(order, mesh, x_deriv)
       >>> wl = 1.5
       >>> (isplines_total3.dItotal_dweights(weights, wl).shape ==
       ...  (len(x_deriv), len(weights)))
       True
       >>> weightslist = list(weights)
       >>> for ix, iw in itertools.product(range(len(x_deriv)),
       ...                                 range(len(weights))):
       ...     w = numpy.array([weightslist[iw]])
       ...     def func(w):
       ...         iweights = numpy.array(weightslist[: iw] +
       ...                                list(w) +
       ...                                weightslist[iw + 1:])
       ...         return isplines_total3.Itotal(iweights, wl)[ix]
       ...     def dfunc(w):
       ...         iweights = numpy.array(weightslist[: iw] +
       ...                                list(w) +
       ...                                weightslist[iw + 1:])
       ...         return isplines_total3.dItotal_dweights(iweights, wl)[ix,
       ...                                                               iw]
       ...     err = scipy.optimize.check_grad(func, dfunc, w)
       ...     if err > 1e-6:
       ...         raise ValueError(f"excess err {err} for {ix, iw}")

       Plot the total of the I-spline family shown in Fig. 1 of
       `Ramsay (1988)`_, adding some linear extrapolation outside the
       mesh range:

       >>> xplot = numpy.linspace(-0.2, 1.2, 1000)
       >>> isplines_totalplot = Isplines_total(order, mesh, xplot)
       >>> df = pd.DataFrame({'x': xplot,
       ...                    'Itotal': isplines_totalplot.Itotal(weights, 0)})
       >>> _ = df.plot(x='x', y='Itotal')

    .. _`Ramsay (1988)`: https://www.jstor.org/stable/2245395

    """

    def __init__(self, order, mesh, x):
        """See main class docstring."""
        if not (isinstance(order, int) and order >= 1):
            raise ValueError(f"`order` not int >= 1: {order}")
        self.order = order

        self.mesh = numpy.array(mesh, dtype='float')
        if self.mesh.ndim != 1:
            raise ValueError(f"`mesh` not array-like of dimension 1: {mesh}")
        if len(self.mesh) < 2:
            raise ValueError(f"`mesh` not length >= 2: {mesh}")
        if not numpy.array_equal(self.mesh, numpy.unique(self.mesh)):
            raise ValueError(f"`mesh` elements not unique and sorted: {mesh}")
        self.lower = self.mesh[0]
        self.upper = self.mesh[-1]
        assert self.lower < self.upper

        self.n = len(self.mesh) - 2 + self.order

        self._x = x.copy()
        self._x.flags.writeable = False

        # indices of `x` in, above, or below I-spline range
        self._index = {'lower': numpy.flatnonzero(self.x < self.lower),
                       'upper': numpy.flatnonzero(self.x > self.upper),
                       'in': numpy.flatnonzero((self.x >= self.lower) &
                                               (self.x <= self.upper))
                       }

        # values of x in each range
        self._x_byrange = {rangename: self.x[index] for rangename, index
                           in self._index.items()}

        # Isplines for each range: for lower and upper it is value at bound
        self._isplines = {'in': Isplines(self.order, self.mesh,
                                         self._x_byrange['in']),
                          'lower': Isplines(self.order, self.mesh,
                                            numpy.array([self.lower])),
                          'upper': Isplines(self.order, self.mesh,
                                            numpy.array([self.upper])),
                          }

        # for caching values
        self._cache = {}
        self._max_cache_size = 100

    @property
    def x(self):
        """numpy.ndarray: Points at which spline is evaluated."""
        return self._x

    def Itotal(self, weights, w_lower):
        r"""Weighted sum of spline family at points :attr:`Isplines_total.x`.

        Parameters
        ----------
        weights : array-like
            Nonnegative weights :math:`w_i` of members :math:`I_i` of spline
            family, should be of length equal to :attr:`Isplines.n`.
        w_lower : float
            The value at the lower bound :math:`L` of the spline range,
            :math:`w_{\rm{lower}}`.

        Returns
        -------
        numpy.ndarray
            :math:`I_{\rm{total}}` for each point in :attr:`Isplines_total.x`.

        """
        args = (tuple(weights), w_lower, 'Itotal')
        if args not in self._cache:
            if len(self._cache) > self._max_cache_size:
                self._cache = {}
            self._cache[args] = self._calculate_Itotal_or_dItotal(*args)
        return self._cache[args]

    def _calculate_Itotal_or_dItotal(self, weights, w_lower, quantity):
        """Calculate :meth:`Isplines.Itotal` or derivatives.

        Parameters have same meaning as for :meth:`Isplines.Itotal`
        except for `quantity`, which should be

          - 'Itotal' to compute :meth:`Isplines.Itotal`
          - 'dItotal_dx' to compute :meth:`Isplines.dItotal_dx`
          - 'dItotal_dweights` to compute :meth:`Isplines.dItotal_dweights`

        Also, `weights` must be hashable (e.g., a tuple).

        """
        # check validity of `weights`
        if len(weights) != self.n:
            raise ValueError(f"invalid length of `weights`: {weights}")
        if any(weight < 0 for weight in weights):
            raise ValueError(f"`weights` not all non-negative: {weights}")

        # compute return values for each category of indices
        returnvals = {}

        if quantity == 'Itotal':
            returnshape = len(self.x)
            if len(self._index['in']):
                returnvals['in'] = numpy.sum([self._isplines['in'].I(i) *
                                              weights[i - 1]
                                              for i in range(1, self.n + 1)],
                                             axis=0) + w_lower
            # values of Itotal at limits
            Itotal_limits = {'lower': w_lower,
                             'upper': w_lower + sum(weights)}
            for name, limit in [('lower', self.lower), ('upper', self.upper)]:
                if not len(self._index[name]):
                    continue
                returnvals[name] = (Itotal_limits[name] +
                                    (self._x_byrange[name] - limit) *
                                    sum(self._isplines[name].dI_dx(i) *
                                        weights[i - 1]
                                        for i in range(1, self.n + 1))
                                    )

        elif quantity == 'dItotal_dx':
            returnshape = len(self.x)
            if len(self._index['in']):
                returnvals['in'] = numpy.sum([self._isplines['in'].dI_dx(i) *
                                              weights[i - 1]
                                              for i in range(1, self.n + 1)],
                                             axis=0)
            for name in ['lower', 'upper']:
                if not len(self._index[name]):
                    continue
                returnvals[name] = sum(self._isplines[name].dI_dx(i) *
                                       weights[i - 1]
                                       for i in range(1, self.n + 1))

        elif quantity == 'dItotal_dweights':
            returnshape = (len(self.x), len(weights))
            if len(self._index['in']):
                returnvals['in'] = (numpy.vstack([self._isplines['in'].I(i) for
                                                  i in range(1, self.n + 1)])
                                    ).transpose()
            # values of I at limits
            I_limits = {'lower': 0.0, 'upper': 1.0}
            for name, limit in [('lower', self.lower), ('upper', self.upper)]:
                if not len(self._index[name]):
                    continue
                returnvals[name] = numpy.vstack(
                                    [I_limits[name] +
                                     (self._x_byrange[name] - limit) *
                                     self._isplines[name].dI_dx(i)
                                     for i in range(1, self.n + 1)
                                     ]
                                    ).transpose()

        else:
            raise ValueError(f"invalid `quantity` {quantity}")

        # reconstruct single return value from indices and returnvalues
        returnval = numpy.full(returnshape, fill_value=numpy.nan)
        for name, name_index in self._index.items():
            if len(name_index):
                returnval[name_index] = returnvals[name]
        assert not numpy.isnan(returnval).any()
        returnval.flags.writeable = False
        return returnval

    def dItotal_dx(self, weights):
        r"""Deriv :meth:`Isplines_total.Itotal` by :attr:`Isplines_total.x`.

        Note
        ----
        Derivatives calculated from equations in :meth:`Isplines_total.Itotal`:

        .. math::

           \frac{\partial I_{\rm{total}}\left(x\right)}{\partial x}
           =
           \begin{cases}
           \sum_i w_i \frac{\partial I_i\left(x\right)}{\partial x}
             & \rm{if\;} L \le x \le U, \\
           \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                      {\partial y}\right\rvert_{y=L}
             & \rm{if\;} x < L, \\
           \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                      {\partial y}\right\rvert_{y=U}
             & \rm{otherwise}.
           \end{cases}

        Note that

        .. math::

           \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                      {\partial y}\right\rvert_{y=L}
            &=&
            \sum_i w_i \left.\frac{\partial I_i\left(y\right)}{\partial y}
                       \right\rvert_{y=L}
            \\
           \left.\frac{\partial I_{\rm{total}}\left(y\right)}
                      {\partial y}\right\rvert_{y=U}
            &=&
            \sum_i w_i \left.\frac{\partial I_i\left(y\right)}{\partial y}
                       \right\rvert_{y=U}

        Parameters
        ----------
        weights : array-like
            Same meaning as for :meth:`Isplines_total.Itotal`.

        Returns
        -------
        numpy.ndarray
            Derivative :math:`\frac{\partial I_{\rm{total}}}{\partial x}`
            for each point in :attr:`Isplines_total.x`.

        """
        args = (tuple(weights), None, 'dItotal_dx')
        if args not in self._cache:
            if len(self._cache) > self._max_cache_size:
                self._cache = {}
            self._cache[args] = self._calculate_Itotal_or_dItotal(*args)
        return self._cache[args]

    def dItotal_dweights(self, weights, w_lower):
        r"""Derivative of :meth:`Isplines_total.Itotal` by :math:`w_i`.

        Parameters
        ----------
        weights : array-like
            Same meaning as for :meth:`Isplines.Itotal`.
        w_lower : float
            Same meaning as for :meth:`Isplines.Itotal`.

        Returns
        -------
        numpy.ndarray
            Array is of shape `(len(x), len(weights))`, and element
            `ix, iweight` gives derivative with respect to weight
            `weights[iweight]` at element `[ix]` of :attr:`Isplines_total.x`.

        Note
        ----
        The derivative is:

        .. math::

           \frac{\partial I_{\rm{total}}\left(x\right)}{\partial w_i}
           =
           \begin{cases}
           I_i\left(x\right)
            & \rm{if\;} L \le x \le U, \\
           I_i\left(L\right) + \left(x-L\right)
           \left.\frac{\partial I_i\left(y\right)}{\partial y}\right\vert_{y=L}
            & \rm{if\;} x < L, \\
           I_i\left(U\right) + \left(x-U\right)
           \left.\frac{\partial I_i\left(y\right)}{\partial y}\right\vert_{y=U}
            & \rm{if\;} x > U.
           \end{cases}

        Note that:

        .. math::

           I_i\left(L\right) &=& 0 \\
           I_i\left(U\right) &=& 1.

        """
        return self._calculate_Itotal_or_dItotal(tuple(weights), w_lower,
                                                 'dItotal_dweights')

    def dItotal_dw_lower(self):
        r"""Deriv of :meth:`Isplines_total.Itotal` by :math:`w_{\rm{lower}}`.

        Returns
        -------
        numpy.ndarray
            :math:`\frac{\partial{I_{\rm{total}}}}{\partial w_{\rm{lower}}}`,
            which is just one for all :attr:`Isplines_total.x`.

        """
        res = numpy.ones(self.x.shape, dtype='float')
        res.flags.writeable = False
        return res


class Isplines:
    r"""Implements I-splines (see `Ramsay (1988)`_).

    Parameters
    ----------
    order : int
        Sets :attr:`Isplines.order`.
    mesh : array-like
        Sets :attr:`Isplines.mesh`.
    x : numpy.ndarray
        Sets :attr:`Isplines.x`.

    Attributes
    ----------
    order : int
        Order of spline, :math:`k` in notation of `Ramsay (1988)`_. Note that
        the degree of the I-spline is equal to :math:`k`, while the
        associated M-spline has order :math:`k` but degree :math:`k - 1`.
    mesh : numpy.ndarray
        Mesh sequence, :math:`\xi_1 < \ldots < \xi_q` in the notation
        of `Ramsay (1988)`_. This class implements **fixed** mesh sequences.
    n : int
        Number of members in spline, denoted as :math:`n` in `Ramsay (1988)`_.
        Related to number of points :math:`q` in the mesh and the order
        :math:`k` by :math:`n = q - 2 + k`.
    lower : float
        Lower end of interval spanned by the splines (first point in mesh).
    upper : float
        Upper end of interval spanned by the splines (last point in mesh).

    Note
    ----
    The methods of this class cache their results and return immutable
    numpy arrays. Do **not** make these arrays mutable and change their
    values, as this will lead to invalid caching.

    Example
    -------
    Short examples to demonstrate and test :class:`Isplines`:

    .. plot::
       :context: reset

       >>> import itertools
       >>> import numpy
       >>> import pandas as pd
       >>> import scipy.optimize
       >>> from dms_variants.ispline import Isplines

       >>> order = 3
       >>> mesh = [0.0, 0.3, 0.5, 0.6, 1.0]
       >>> x = numpy.array([0, 0.2, 0.3, 0.4, 0.8, 0.99999])
       >>> isplines = Isplines(order, mesh, x)
       >>> isplines.order
       3
       >>> isplines.mesh
       array([0. , 0.3, 0.5, 0.6, 1. ])
       >>> isplines.n
       6
       >>> isplines.lower
       0.0
       >>> isplines.upper
       1.0

       Evaluate the I-splines at some selected points:

       >>> for i in range(1, isplines.n + 1):
       ...     print(f"I{i}: {numpy.round(isplines.I(i), 2)}")
       ... # doctest: +NORMALIZE_WHITESPACE
       I1: [0.   0.96 1.   1.   1.   1.  ]
       I2: [0.   0.52 0.84 0.98 1.   1.  ]
       I3: [0.   0.09 0.3  0.66 1.   1.  ]
       I4: [0.   0.   0.   0.02 0.94 1.  ]
       I5: [0.   0.   0.   0.   0.58 1.  ]
       I6: [0.   0.   0.   0.   0.13 1.  ]

       Check that gradients are correct for :meth:`Isplines.dI_dx`:

       >>> for i, xval in itertools.product(range(1, isplines.n + 1), x):
       ...     xval = numpy.array([xval])
       ...     def func(xval):
       ...         return Isplines(order, mesh, xval).I(i)
       ...     def dfunc(xval):
       ...         return Isplines(order, mesh, xval).dI_dx(i)
       ...     err = scipy.optimize.check_grad(func, dfunc, xval)
       ...     if err > 1e-5:
       ...         raise ValueError(f"excess err {err} for {i}, {xval}")

       Plot the I-splines in Fig. 1 of `Ramsay (1988)`_:

       >>> xplot = numpy.linspace(0, 1, 1000)
       >>> isplines_xplot = Isplines(order, mesh, xplot)
       >>> data = {'x': xplot}
       >>> for i in range(1, isplines.n + 1):
       ...     data[f"I{i}"] = isplines_xplot.I(i)
       >>> df = pd.DataFrame(data)
       >>> _ = df.plot(x='x')

    .. _`Ramsay (1988)`: https://www.jstor.org/stable/2245395

    """

    def __init__(self, order, mesh, x):
        """See main class docstring."""
        if not (isinstance(order, int) and order >= 1):
            raise ValueError(f"`order` not int >= 1: {order}")
        self.order = order

        self.mesh = numpy.array(mesh, dtype='float')
        if self.mesh.ndim != 1:
            raise ValueError(f"`mesh` not array-like of dimension 1: {mesh}")
        if len(self.mesh) < 2:
            raise ValueError(f"`mesh` not length >= 2: {mesh}")
        if not numpy.array_equal(self.mesh, numpy.unique(self.mesh)):
            raise ValueError(f"`mesh` elements not unique and sorted: {mesh}")
        self.lower = self.mesh[0]
        self.upper = self.mesh[-1]
        assert self.lower < self.upper

        self.n = len(self.mesh) - 2 + self.order

        if not (isinstance(x, numpy.ndarray) and x.ndim == 1):
            raise ValueError('`x` is not numpy.ndarray of dimension 1')
        if (x < self.lower).any() or (x > self.upper).any():
            raise ValueError(f"`x` outside {self.lower} and {self.upper}: {x}")
        self._x = x.copy()
        self._x.flags.writeable = False

        self._msplines = Msplines(order + 1, mesh, self.x)

        # for caching values
        self._cache = {}
        self._max_cache_size = 100

    @property
    def x(self):
        """numpy.ndarray: Points at which spline is evaluated."""
        return self._x

    def I(self, i):  # noqa: E743,E741
        r"""Evaluate spline :math:`I_i` at point(s) :attr:`Isplines.x`.

        Parameters
        ----------
        i : int
            Spline member :math:`I_i`, where :math:`1 \le i \le`
            :attr:`Isplines.n`.

        Returns
        -------
        numpy.ndarray
            The values of the I-spline at each point in :attr:`Isplines.x`.

        Note
        ----
        The spline is evaluated using the formula given in the
        `Praat manual`_, which corrects some errors in the formula
        provided by `Ramsay (1988)`_:

        .. math::

           I_i\left(x\right)
           =
           \begin{cases}
           0 & \rm{if\;} i > j, \\
           1 & \rm{if\;} i < j - k, \\
           \sum_{m=i+1}^j \left(t_{m+k+1} - t_m\right)
                          M_m\left(x \mid k + 1\right) / \left(k + 1 \right)
             & \rm{otherwise},
           \end{cases}

        where :math:`j` is the index such that :math:`t_j \le x < t_{j+1}`
        (the :math:`\left\{t_j\right\}` are the :attr:`Msplines.knots` for a
        M-spline of order :math:`k + 1`) and :math:`k` is
        :attr:`Isplines.order`.

        .. _`Ramsay (1988)`: https://www.jstor.org/stable/2245395
        .. _`Praat manual`: http://www.fon.hum.uva.nl/praat/manual/spline.html

        """
        args = (i, 'I')
        if args not in self._cache:
            if len(self._cache) > self._max_cache_size:
                self._cache = {}
            self._cache[args] = self._calculate_I_or_dI(*args)
        return self._cache[args]

    @property
    def j(self):
        """numpy.ndarray: :math:`j` as defined in :meth:`Isplines.I`."""
        if not hasattr(self, '_j'):
            self._j = numpy.searchsorted(self._msplines.knots, self.x, 'right')
            assert ((1 <= self._j).all() and
                    (self._j <= len(self._msplines.knots)).all())
            assert self.x.shape == self._j.shape
        return self._j

    @property
    def _sum_terms_I(self):
        """numpy.ndarray: sum terms for :meth:`Isplines.I`.

        Row `m - 1` has summation term for `m`.

        """
        if not hasattr(self, '_sum_terms_I_val'):
            k = self.order
            self._sum_terms_I_val = numpy.vstack(
                [(self._msplines.knots[m + k] - self._msplines.knots[m - 1]) *
                 self._msplines.M(m, k + 1) / (k + 1)
                 for m in range(1, self._msplines.n + 1)])
            assert self._sum_terms_I_val.shape == (self._msplines.n,
                                                   len(self.x))
        return self._sum_terms_I_val

    @property
    def _sum_terms_dI_dx(self):
        """numpy.ndarray: sum terms for :meth:`Isplines.dI_dx`.

        Row `m - 1` has summation term for `m`.

        """
        if not hasattr(self, '_sum_terms_dI_dx_val'):
            k = self.order
            self._sum_terms_dI_dx_val = numpy.vstack(
                [(self._msplines.knots[m + k] - self._msplines.knots[m - 1]) *
                 self._msplines.dM_dx(m, k + 1) / (k + 1)
                 for m in range(1, self._msplines.n + 1)])
            assert self._sum_terms_dI_dx_val.shape == (self._msplines.n,
                                                       len(self.x))
        return self._sum_terms_dI_dx_val

    def _calculate_I_or_dI(self, i, quantity):
        """Calculate :meth:`Isplines.I` or :meth:`Isplines.dI_dx`.

        Parameters
        ----------
        i : int
            Same meaning as for :meth:`Isplines.I`.
        quantity : {'I', 'dI'}
            Calculate :meth:`Isplines.I` or :meth:`Isplines.dI_dx`?

        Returns
        -------
        numpy.ndarray
            The return value of :meth:`Isplines.I` or :meth:`Isplines.dI_dx`.

        Note
        ----
        Most calculations for :meth:`Isplines.I` and :meth:`Isplines.dI_dx`
        are the same, so this method implements both.

        """
        if quantity == 'I':
            sum_terms = self._sum_terms_I
            i_lt_jminusk = 1.0
        elif quantity == 'dI':
            sum_terms = self._sum_terms_dI_dx
            i_lt_jminusk = 0.0
        else:
            raise ValueError(f"invalid `quantity` {quantity}")

        if not (1 <= i <= self.n):
            raise ValueError(f"invalid spline member `i` of {i}")

        k = self.order

        # create `binary_terms` where entry (m - 1, x) is 1 if and only if
        # the corresponding `sum_terms` entry is part of the sum.
        binary_terms = numpy.vstack(
               [numpy.zeros(len(self.x)) if m < i + 1
                else (m <= self.j).astype(int)
                for m in range(1, self._msplines.n + 1)])
        assert binary_terms.shape == sum_terms.shape

        # compute sums from `sum_terms` and `binary_terms`
        sums = numpy.sum(sum_terms * binary_terms, axis=0)
        assert sums.shape == self.x.shape

        # return value with sums, 0, or 1
        res = numpy.where(i > self.j, 0.0,
                          numpy.where(i < self.j - k, i_lt_jminusk,
                                      sums))
        res.flags.writeable = False
        return res

    def dI_dx(self, i):
        r"""Derivative of :meth:`Isplines.I` by :attr:`Isplines.x`.

        Parameters
        ----------
        i : int
            Same meaning as for :meth:`Isplines.I`.

        Returns
        -------
        numpy.ndarray
            Derivative of I-spline with respect to :attr:`Isplines.x`.

        Note
        ----
        The derivative is calculated from the equation in :meth:`Isplines.I`:

        .. math::

           \frac{\partial I_i\left(x\right)}{\partial x}
           =
           \begin{cases}
           0 & \rm{if\;} i > j \rm{\; or \;} i < j - k, \\
           \sum_{m=i+1}^j\left(t_{m+k+1} - t_m\right)
                         \frac{\partial M_m\left(x \mid k+1\right)}{\partial x}
                         \frac{1}{k + 1}
             & \rm{otherwise}.
           \end{cases}

        """
        args = (i, 'dI')
        if args not in self._cache:
            if len(self._cache) > self._max_cache_size:
                self._cache = {}
            self._cache[args] = self._calculate_I_or_dI(*args)
        return self._cache[args]


class Msplines:
    r"""Implements M-splines (see `Ramsay (1988)`_).

    Parameters
    ----------
    order : int
        Sets :attr:`Msplines.order`.
    mesh : array-like
        Sets :attr:`Msplines.mesh`.
    x : numpy.ndarray
        Sets :attr:`Msplines.x`.

    Attributes
    ----------
    order : int
        Order of spline, :math:`k` in notation of `Ramsay (1988)`_.
        Polynomials are of degree :math:`k - 1`.
    mesh : numpy.ndarray
        Mesh sequence, :math:`\xi_1 < \ldots < \xi_q` in the notation
        of `Ramsay (1988)`_. This class implements **fixed** mesh sequences.
    n : int
        Number of members in spline, denoted as :math:`n` in `Ramsay (1988)`_.
        Related to number of points :math:`q` in the mesh and the order
        :math:`k` by :math:`n = q - 2 + k`.
    knots : numpy.ndarray
        The knot sequence, :math:`t_1, \ldots, t_{n + k}` in the notation of
        `Ramsay (1988)`_.
    lower : float
        Lower end of interval spanned by the splines (first point in mesh).
    upper : float
        Upper end of interval spanned by the splines (last point in mesh).

    Note
    ----
    The methods of this class cache their results and return immutable
    numpy arrays. Do **not** make those arrays mutable and change their
    values as this will lead to invalid caching.

    Example
    -------
    Demonstrate and test :class:`Msplines`:

    .. plot::
       :context: reset

       >>> import functools
       >>> import itertools
       >>> import numpy
       >>> import pandas as pd
       >>> import scipy.optimize
       >>> from dms_variants.ispline import Msplines

       >>> order = 3
       >>> mesh = [0.0, 0.3, 0.5, 0.6, 1.0]
       >>> x = numpy.array([0, 0.2, 0.3, 0.4, 0.8, 0.99999])
       >>> msplines = Msplines(order, mesh, x)
       >>> msplines.order
       3
       >>> msplines.mesh
       array([0. , 0.3, 0.5, 0.6, 1. ])
       >>> msplines.n
       6
       >>> msplines.knots
       array([0. , 0. , 0. , 0.3, 0.5, 0.6, 1. , 1. , 1. ])
       >>> msplines.lower
       0.0
       >>> msplines.upper
       1.0

       Evaluate the M-splines at some selected points:

       >>> for i in range(1, msplines.n + 1):
       ...     print(f"M{i}: {numpy.round(msplines.M(i), 2)}")
       ... # doctest: +NORMALIZE_WHITESPACE
       M1: [10. 1.11 0.  0.   0.   0.  ]
       M2: [0.  3.73 2.4 0.6  0.   0.  ]
       M3: [0.  1.33 3.  3.67 0.   0.  ]
       M4: [0.  0.   0.  0.71 0.86 0.  ]
       M5: [0.  0.   0.  0.   3.3  0.  ]
       M6: [0.  0.   0.  0.   1.88 7.5 ]

       Check that the gradients are correct:

       >>> for i, xval in itertools.product(range(1, msplines.n + 1), x):
       ...     xval = numpy.array([xval])
       ...     def func(xval):
       ...         return Msplines(order, mesh, xval).M(i)
       ...     def dfunc(xval):
       ...         return Msplines(order, mesh, xval).dM_dx(i)
       ...     err = scipy.optimize.check_grad(func, dfunc, xval)
       ...     if err > 1e-5:
       ...         raise ValueError(f"excess err {err} for {i}, {xval}")

       Plot the M-splines in in Fig. 1 of `Ramsay (1988)`_:

       >>> xplot = numpy.linspace(0, 1, 1000, endpoint=False)
       >>> msplines_plot = Msplines(order, mesh, xplot)
       >>> data = {'x': xplot}
       >>> for i in range(1, msplines_plot.n + 1):
       ...     data[f"M{i}"] = msplines_plot.M(i)
       >>> df = pd.DataFrame(data)
       >>> _ = df.plot(x='x')

    .. _`Ramsay (1988)`: https://www.jstor.org/stable/2245395

    """

    def __init__(self, order, mesh, x):
        """See main class docstring."""
        if not (isinstance(order, int) and order >= 1):
            raise ValueError(f"`order` not int >= 1: {order}")
        self.order = order

        self.mesh = numpy.array(mesh, dtype='float')
        if self.mesh.ndim != 1:
            raise ValueError(f"`mesh` not array-like of dimension 1: {mesh}")
        if len(self.mesh) < 2:
            raise ValueError(f"`mesh` not length >= 2: {mesh}")
        if not numpy.array_equal(self.mesh, numpy.unique(self.mesh)):
            raise ValueError(f"`mesh` elements not unique and sorted: {mesh}")
        self.lower = self.mesh[0]
        self.upper = self.mesh[-1]
        assert self.lower < self.upper

        self.knots = numpy.array(
                        [self.lower] * self.order +
                        list(self.mesh[1: -1]) +
                        [self.upper] * self.order,
                        dtype='float')

        self.n = len(self.knots) - self.order
        assert self.n == len(self.mesh) - 2 + self.order

        if not (isinstance(x, numpy.ndarray) and x.ndim == 1):
            raise ValueError('`x` is not numpy.ndarray of dimension 1')
        if (x < self.lower).any() or (x > self.upper).any():
            raise ValueError(f"`x` outside {self.lower} and {self.upper}: {x}")
        self._x = x.copy()
        self._x.flags.writeable = False

        self._ti_le_x_lt_tiplusk_cache = {}

        # for caching values
        self._M_cache = {}
        self._dM_cache = {}
        self._max_cache_size = 100

    def _ti_le_x_lt_tiplusk(self, ti, tiplusk):
        r"""Indices where :math:`t_i \le x \le t_{i+k}`.

        Parameters
        ----------
        ti : float
            :math:`t_i`
        tiplusk : float
            :math:`t_{i+k}`

        Returns
        -------
        numpy.ndarray
            Array of booleans of same length as :attr:`Msplines.x` indicating
            if :math:`t_i \le x \le t_{i+k}`.

        """
        key = (ti, tiplusk)
        if key not in self._ti_le_x_lt_tiplusk_cache:
            val = (ti <= self.x) & (self.x < tiplusk)
            val.flags.writeable = False
            assert val.dtype == bool
            if len(self._ti_le_x_lt_tiplusk_cache) > self._max_cache_size:
                self._ti_le_x_lt_tiplusk_cache = {}
            self._ti_le_x_lt_tiplusk_cache[key] = val
        return self._ti_le_x_lt_tiplusk_cache[key]

    @property
    def x(self):
        """numpy.ndarray: Points at which spline is evaluated."""
        return self._x

    def M(self, i, k=None, invalid_i='raise'):
        r"""Evaluate spline :math:`M_i` at point(s) :attr:`Msplines.x`.

        Parameters
        ----------
        i : int
            Spline member :math:`M_i`, where :math:`1 \le i \le`
            :attr:`Msplines.n`.
        k : int or None
            Order of spline. If `None`, assumed to be :attr:`Msplines.order`.
        invalid_i : {'raise', 'zero'}
            If `i` is invalid, do we raise an error or return 0?

        Returns
        -------
        numpy.ndarray
            The values of the M-spline at each point in :attr:`Msplines.x`.

        Note
        ----
        The spline is evaluated using the recursive relationship given by
        `Ramsay (1988) <https://www.jstor.org/stable/2245395>`_:

        .. math::

           M_i\left(x \mid k=1\right)
           &=&
           \begin{cases}
           1 / \left(t_{i+1} - t_i\right), & \rm{if\;} t_i \le x < t_{i+1} \\
           0, & \rm{otherwise}
           \end{cases} \\
           M_i\left(x \mid k > 1\right) &=&
           \begin{cases}
           \frac{k\left[\left(x - t_i\right) M_i\left(x \mid k-1\right) +
                        \left(t_{i+k} -x\right) M_{i+1}\left(x \mid k-1\right)
                        \right]}
           {\left(k - 1\right)\left(t_{i + k} - t_i\right)},
           & \rm{if\;} t_i \le x < t_{i+k} \\
           0, & \rm{otherwise}
           \end{cases}

        """
        args = (i, k, invalid_i)
        if args not in self._M_cache:
            if len(self._M_cache) > self._max_cache_size:
                self._M_cache = {}
            self._M_cache[args] = self._calculate_M(*args)
        return self._M_cache[args]

    def _calculate_M(self, i, k, invalid_i):
        """Calculate :meth:`Msplines.M` with result caching."""
        if not (1 <= i <= self.n):
            if invalid_i == 'raise':
                raise ValueError(f"invalid spline member `i` of {i}")
            elif invalid_i == 'zero':
                return 0
            else:
                raise ValueError(f"invalid `invalid_i` of {invalid_i}")
        if k is None:
            k = self.order
        if not 1 <= k <= self.order:
            raise ValueError(f"invalid spline order `k` of {k}")

        tiplusk = self.knots[i + k - 1]
        ti = self.knots[i - 1]
        if tiplusk == ti:
            return 0

        boolindex = self._ti_le_x_lt_tiplusk(ti, tiplusk)
        if k == 1:
            res = numpy.where(boolindex, 1.0 / (tiplusk - ti), 0.0)
            res.flags.writeable = False
            return res
        else:
            assert k > 1
            res = numpy.where(
                        boolindex,
                        (k * ((self.x - ti) * self.M(i, k - 1) +
                         (tiplusk - self.x) * self.M(i + 1, k - 1,
                                                     invalid_i='zero')
                         ) / ((k - 1) * (tiplusk - ti))),
                        0.0)
            res.flags.writeable = False
            return res

    def dM_dx(self, i, k=None, invalid_i='raise'):
        r"""Derivative of :meth:`Msplines.M` by to :attr:`Msplines.x`.

        Parameters
        ----------
        i : int
            Same as for :meth:`Msplines.M`.
        k : int or None
            Same as for :meth:`Msplines.M`.
        invalid_i : {'raise', 'zero'}
            Same as for :meth:`Msplines.M`.

        Returns
        -------
        numpy.ndarray
            Derivative of M-spline with respect to :attr:`Msplines.x`.

        Note
        ----
        The derivative is calculated from the equation in :meth:`Msplines.M`:

        .. math::

           \frac{\partial M_i\left(x \mid k=1\right)}{\partial x} &=& 0
           \\
           \frac{\partial M_i\left(x \mid k > 1\right)}{\partial x}
           &=&
           \begin{cases}
           \frac{k\left[\left(x - t_i\right)
                        \frac{\partial M_i\left(x \mid k-1\right)}{\partial x}
                        +
                        M_i\left(x \mid k-1\right)
                        +
                        \left(t_{i+k} -x\right)
                        \frac{\partial M_{i+1}\left(x \mid k-1\right)}
                             {\partial x}
                        -
                        M_{i+1}\left(x \mid k-1\right)
                        \right]}
           {\left(k - 1\right)\left(t_{i + k} - t_i\right)},
           & \rm{if\;} t_i \le x < t_{i+1} \\
           0, & \rm{otherwise}
           \end{cases}

        """
        args = (i, k, invalid_i)
        if args not in self._dM_cache:
            if len(self._dM_cache) > self._max_cache_size:
                self._dM_cache = {}
            self._dM_cache[args] = self._calculate_dM_dx(*args)
        return self._dM_cache[args]

    def _calculate_dM_dx(self, i, k, invalid_i):
        """Calculate :meth:`Msplines.dM_dx` with results caching."""
        if not (1 <= i <= self.n):
            if invalid_i == 'raise':
                raise ValueError(f"invalid spline member `i` of {i}")
            elif invalid_i == 'zero':
                return 0
            else:
                raise ValueError(f"invalid `invalid_i` of {invalid_i}")
        if k is None:
            k = self.order
        if not 1 <= k <= self.order:
            raise ValueError(f"invalid spline order `k` of {k}")

        tiplusk = self.knots[i + k - 1]
        ti = self.knots[i - 1]
        if tiplusk == ti or k == 1:
            return 0
        else:
            assert k > 1
            boolindex = self._ti_le_x_lt_tiplusk(ti, tiplusk)
            res = numpy.where(
                        boolindex,
                        (k * ((self.x - ti) * self.dM_dx(i, k - 1) +
                              self.M(i, k - 1) +
                              (tiplusk - self.x) * self.dM_dx(i + 1, k - 1,
                                                              invalid_i='zero')
                              - self.M(i + 1, k - 1, invalid_i='zero')
                              ) / ((k - 1) * (tiplusk - ti))
                         ),
                        0.0)
            res.flags.writeable = False
            return res


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
===========
utils
===========

Miscellaneous utility functions.

"""


import re

import matplotlib.ticker

import pandas as pd  # noqa: F401

import dms_variants._cutils
from dms_variants.constants import (AAS_NOSTOP,
                                    CODON_TO_AA,
                                    NT_COMPLEMENT,
                                    )


def reverse_complement(s, *, use_cutils=True):
    """Get reverse complement of DNA sequence.

    Parameters
    ----------
    s : str
        DNA sequence.
    use_cutils : bool
        Use faster C-extension implementation.

    Returns
    -------
    str
        Reverse complement of `s`.

    Example
    -------
    >>> s = 'ATGCAAN'
    >>> reverse_complement(s)
    'NTTGCAT'
    >>> reverse_complement(s, use_cutils=False) == reverse_complement(s)
    True

    """
    if use_cutils:
        return dms_variants._cutils.reverse_complement(s)
    else:
        return ''.join(reversed([NT_COMPLEMENT[nt] for nt in s]))


def translate(codonseq):
    """Translate codon sequence.

    Parameters
    ----------
    codonseq : str
        Codon sequence. Gaps currently not allowed.

    Returns
    -------
    str
        Amino-acid sequence.

    Example
    -------
    >>> translate('ATGGGATAA')
    'MG*'

    """
    if len(codonseq) % 3 != 0:
        raise ValueError('length of `codonseq` not multiple of 3')

    aaseq = []
    for icodon in range(len(codonseq) // 3):
        codon = codonseq[3 * icodon: 3 * icodon + 3]
        aaseq.append(CODON_TO_AA[codon])

    return ''.join(aaseq)


def latex_sci_not(xs):
    r"""Convert list of numbers to LaTex scientific notation.

    Parameters
    ----------
    xs : list
        Numbers to format.

    Returns
    -------
    list
        Formatted strings for numbers.

    Examples
    ----------
    >>> latex_sci_not([0, 3, 3120, -0.0000927])
    ['$0$', '$3$', '$3.1 \\times 10^{3}$', '$-9.3 \\times 10^{-5}$']

    >>> latex_sci_not([0.001, 1, 1000, 1e6])
    ['$0.001$', '$1$', '$10^{3}$', '$10^{6}$']

    >>> latex_sci_not([-0.002, 0.003, 0.000011])
    ['$-0.002$', '$0.003$', '$1.1 \\times 10^{-5}$']

    >>> latex_sci_not([-0.1, 0.0, 0.1, 0.2])
    ['$-0.1$', '$0$', '$0.1$', '$0.2$']

    >>> latex_sci_not([0, 1, 2])
    ['$0$', '$1$', '$2$']

    """
    formatlist = []
    for x in xs:
        xf = f"{x:.2g}"
        if xf[: 2] == '1e':
            xf = f"$10^{{{int(xf[2 : ])}}}$"
        elif xf[: 3] == '-1e':
            xf = f"$-10^{{{int(xf[3 : ])}}}$"
        elif 'e' in xf:
            d, exp = xf.split('e')
            xf = f"${d} \\times 10^{{{int(exp)}}}$"
        else:
            xf = f"${xf}$"
        formatlist.append(xf)
    return formatlist


def cumul_rows_by_count(df, *, count_col='count', n_col='n_rows',
                        tot_col='total_rows', group_cols=None,
                        group_cols_as_str=False):
    """Cumulative number of rows with >= each count.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with rows to analyze.
    count_col : str
        Column in `df` with count for row.
    n_col : str
        Name of column in result giving cumulative count threshold.
    tot_col : str
        Name of column in result giving total number of rows.
    group_cols : None or list
        Group by these columns and analyze each group separately.
    group_cols_as_str : bool
        Convert any `group_cols` columns to str. This is needed if calling
        in ``R`` using `reticulate <https://rstudio.github.io/reticulate/>`_.

    Returns
    -------
    pandas.DataFrame
        Give cumulative counts. For each count in `count_col`, column
        `n_col` gives number of rows with >= that many counts, and
        `tot_col` gives total number of counts.

    Examples
    --------
    >>> df = pd.DataFrame({'sample': ['a', 'a', 'b', 'b', 'a', 'a'],
    ...                    'count': [9, 0, 1, 4, 3, 3]})
    >>> cumul_rows_by_count(df)
       count  n_rows  total_rows
    0      9       1           6
    1      4       2           6
    2      3       4           6
    3      1       5           6
    4      0       6           6
    >>> cumul_rows_by_count(df, group_cols=['sample'])
      sample  count  n_rows  total_rows
    0      a      9       1           4
    1      a      3       3           4
    2      a      0       4           4
    3      b      4       1           2
    4      b      1       2           2

    """
    if count_col not in df.columns:
        raise ValueError(f"df does not have column {count_col}")

    if not group_cols:
        drop_group_cols = True
        group_cols = ['dummy_col']
        df[group_cols[0]] = '_'
    else:
        drop_group_cols = False
        if isinstance(group_cols, str):
            group_cols = [group_cols]

    if not set(group_cols).issubset(set(df.columns)):
        raise ValueError(f"invalid `group_cols` {group_cols}")

    for col in [count_col, n_col, tot_col]:
        if col in group_cols:
            raise ValueError(f"`group_cols` cannot contain {col}")

    df = (
         df

         # get number of rows with each count
         .assign(**{n_col: 1})
         .groupby(group_cols + [count_col])
         .aggregate({n_col: 'count'})
         .reset_index()
         .sort_values(group_cols + [count_col],
                      ascending=[True] * len(group_cols) + [False])
         .reset_index(drop=True)

         # get cumulative number with <= number of counts
         .assign(**{n_col: lambda x: x.groupby(group_cols)[n_col].cumsum()})

         # add new column that is total number of rows
         .assign(**{tot_col: lambda x: (x
                                        .groupby(group_cols, sort=False)
                                        [n_col]
                                        .transform('max'))})
         )
    assert df[tot_col].notnull().all()

    if drop_group_cols:
        df = df.drop(group_cols, axis='columns')
    elif group_cols_as_str:
        for col in group_cols:
            df[col] = df[col].astype('str')

    return df


def tidy_to_corr(df, sample_col, label_col, value_col, *,
                 group_cols=None, return_type='tidy_pairs',
                 method='pearson'):
    """Pairwise correlations between samples in tidy data frame.

    Parameters
    ----------
    df : pandas.DataFrame
        Tidy data frame.
    sample_col : str
        Column in `df` with name of sample.
    label_col : str
        Column in `df` with labels for variable to correlate.
    value_col : str
        Column in `df` with values to correlate.
    group_cols : None, str, or list
        Additional columns used to group results.
    return_type : {'tidy_pairs', 'matrix'}
        Return results as tidy dataframe of pairwise correlations
        or correlation matrix.
    method : str
        A correlation metho passable to `pandas.DataFrame.corr`.

    Returns
    -------
    pandas.DataFrame
        Holds pairwise correlations in format specified by `return_type`.
        Correlations only calculated among values with shared label
        among samples.

    Example
    -------
    Define data frame with data to correlate:

    >>> df = pd.DataFrame({
    ...        'sample': ['a', 'a', 'a', 'b', 'b', 'b', 'b', 'c', 'c', 'c'],
    ...        'barcode': ['A', 'C', 'G', 'G', 'A', 'C', 'T', 'G', 'C', 'A'],
    ...        'score': [1, 2, 3, 3, 1.5, 2, 4, 1, 2, 3],
    ...        'group': ['x', 'x', 'x', 'x', 'x', 'x', 'x', 'y', 'y', 'y'],
    ...        })

    Pairwise correlations between all samples ignoring group:

    >>> tidy_to_corr(df, sample_col='sample', label_col='barcode',
    ...              value_col='score')
      sample_1 sample_2  correlation
    0        a        a     1.000000
    1        b        a     0.981981
    2        c        a    -1.000000
    3        a        b     0.981981
    4        b        b     1.000000
    5        c        b    -0.981981
    6        a        c    -1.000000
    7        b        c    -0.981981
    8        c        c     1.000000

    The same but as a matrix rather than in tidy format:

    >>> tidy_to_corr(df, sample_col='sample', label_col='barcode',
    ...              value_col='score', return_type='matrix')
      sample         a         b         c
    0      a  1.000000  0.981981 -1.000000
    1      b  0.981981  1.000000 -0.981981
    2      c -1.000000 -0.981981  1.000000

    Now group before computing correlations:

    >>> tidy_to_corr(df, sample_col='sample', label_col='barcode',
    ...              value_col='score', group_cols='group')
      group sample_1 sample_2  correlation
    0     x        a        a     1.000000
    1     x        b        a     0.981981
    2     x        a        b     0.981981
    3     x        b        b     1.000000
    4     y        c        c     1.000000
    >>> tidy_to_corr(df, sample_col='sample', label_col='barcode',
    ...              value_col='score', group_cols='group',
    ...              return_type='matrix')
      group sample         a         b    c
    0     x      a  1.000000  0.981981  NaN
    1     x      b  0.981981  1.000000  NaN
    2     y      c       NaN       NaN  1.0

    """
    if isinstance(group_cols, str):
        group_cols = [group_cols]
    elif group_cols is None:
        group_cols = []
    cols = [sample_col, value_col, label_col] + group_cols
    if set(cols) > set(df.columns):
        raise ValueError(f"`df` missing some of these columns: {cols}")
    if len(set(cols)) != len(cols):
        raise ValueError(f"duplicate column names: {cols}")
    if 'correlation' in cols:
        raise ValueError('cannot have column named `correlation`')
    if sample_col + '_2' in group_cols:
        raise ValueError(f"cannot have column named `{sample_col}_2`")

    for _, g in df.groupby([sample_col] + group_cols):
        if len(g[label_col]) != g[label_col].nunique():
            raise ValueError(f"Entries in `df` column {label_col} not unique "
                             'after grouping by: ' +
                             ', '.join(c for c in [sample_col] + group_cols))

    df = (
        df
        .pivot_table(values=value_col,
                     columns=sample_col,
                     index=[label_col] + group_cols,
                     )
        .reset_index()
        )

    if group_cols:
        df = df.groupby(group_cols)

    corr = (
        df
        .corr(method=method)
        .dropna(how='all', axis='index')
        .reset_index()
        )

    corr.columns.name = None  # remove name of columns index

    if return_type == 'tidy_pairs':
        corr = (
            corr
            .melt(id_vars=group_cols + [sample_col],
                  var_name=sample_col + '_2',
                  value_name='correlation'
                  )
            .rename(columns={sample_col: sample_col + '_1'})
            .dropna()
            .reset_index(drop=True)
            )

    elif return_type != 'matrix':
        raise ValueError(f"invalid `return_type` of {return_type}")

    return corr


def tidy_split(df, column, sep=' ', keep=False):
    """Split values of column and expand into new rows.

    Note
    ----
    Taken from https://stackoverflow.com/a/39946744

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with the column to split and expand.
    column : str
        Name of column to split and expand.
    sep : str
        The string used to split the column's values.
    keep : bool
        Retain the presplit value as it's own row.

    Returns
    -------
    pandas.DataFrame
        Data frame with the same columns as `df`. Rows lacking
        `column` are filtered.

    Example
    -------
    >>> df = pd.DataFrame({'col1': ['A', 'B', 'C'],
    ...                    'col2': ['d e', float('nan'), 'f']})
    >>> tidy_split(df, 'col2')
      col1 col2
    0    A    d
    0    A    e
    2    C    f

    """
    indexes = []
    new_values = []
    df = df.dropna(subset=[column])
    for i, presplit in enumerate(df[column].astype(str)):
        values = presplit.split(sep)
        if keep and len(values) > 1:
            indexes.append(i)
            new_values.append(presplit)
        for value in values:
            indexes.append(i)
            new_values.append(value)
    new_df = df.iloc[indexes, :].copy()
    new_df[column] = new_values
    return new_df


def integer_breaks(x):
    """Integer breaks for axes labels.

    Note
    ----
    The breaks can be passed to `plotnine <http://plotnine.readthedocs.io>`_
    as in::

        scale_x_continuous(breaks=integer_breaks)

    Parameters
    ----------
    x : array-like
        Numerical data values.

    Returns
    -------
    numpy.ndarray
        Integer tick locations.

    Example
    -------
    >>> integer_breaks([0.5, 0.7, 1.2, 3.7, 7, 17])
    array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])

    """
    return (matplotlib.ticker.MaxNLocator(integer=True)
            .tick_values(min(x), max(x))
            )


def scores_to_prefs(df, mutation_col, score_col, base,
                    wt_score=0, missing='average',
                    alphabet=AAS_NOSTOP, exclude_chars=('*',),
                    returnformat='wide', stringency_param=1):
    r"""Convert functional scores to amino-acid preferences.

    Preferences are calculated from functional scores as follows. Let
    :math:`y_{r,a}` be the score of the variant with the single mutation
    of site :math:`r` to :math:`a` (when :math:`a` is the wildtype character,
    then :math:`p_{r,a}` is the score of the wildtype sequence). Then the
    preference :math:`\pi_{r,a}` is

    .. math::

        \pi_{r,a} = \frac{b^{y_{r,a}}}{\sum_{a'} b^{y_{r,a'}}}

    where :math:`b` is the base for the exponent. This definition ensures
    that the preferences sum to one at each site. These preferences can be
    displayed in logo pltos or used as input to
    `phydms <https://jbloomlab.github.io/phydms/>`_

    Note
    ----
    The "flatness" of the preferences is determined by the exponent base.
    A smaller `base` yields flatter preferences. There is no obvious "best"
    `base` as different values correspond to different linear scalings
    of the scores. A recommended approach is simply to choose a value of `base`
    (such as 10) and then re-scale the preferences by using
    `phydms <https://jbloomlab.github.io/phydms/>`_ to optimize a stringency
    parameter as `described here <https://peerj.com/articles/3657>`_. One
    thing to note is that `phydms <https://jbloomlab.github.io/phydms/>`_
    has an upper bound on the largest stringency parameter it can fit,
    so if you are hitting this upper bound then pre-scale the preferences
    to be less flat by using a larger value of `base`.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame holding the functional scores.
    mutation_col : str
        Column in `df` with mutations, in this format: 'M1A'.
    score_col : str
        Column in `df` with functional scores.
    base : float
        Base to which the exponent is taken in computing the preferences.
        Make sure not to choose an excessively small value if using
        in `phydms <https://jbloomlab.github.io/phydms/>`_ or the
        preferences will be too flat. In the examples below we use 2,
        but you may want a larger value.
    wt_score : float
        Functional score for wildtype sequence.
    missing : {'average', 'site_average', 'error'}
        What to do when there is no estimate of the score for a mutant?
        Estimate the phenotype as the average of all single mutants, the
        average of all single mutants at that site, or raise an error.
    alphabet : list or tuple
        Characters (e.g., amino acids) for which we compute preferences.
    exclude_chars : tuple or list
        Characters to exclude when calculating preferences (and when
        averaging values for missing mutants). For instance, you might
        want to exclude stop codons even if they are in `df`.
    returnformat : {'tidy', 'wide'}
        Return preferences in tidy or wide format data frame.
    stringency_param : float
        Re-scale preferences by this stringency parameter. This
        involves raising each preference to the power of
        `stringency_param`, and then re-normalizes. A similar
        effect can be achieved by changing `base`.

    Returns
    -------
    pandas.DataFrame
        Data frame where first column is named 'site', other columns are
        named for each character, and rows give preferences for each site.

    Example
    -------
    >>> func_scores_df = pd.DataFrame(
    ...         {'aa_substitutions': ['M1A', 'M1C', 'A2M', 'A2C', 'M1*'],
    ...          'func_score':       [-0.1,  -2.3,   0.8,  -1.2,  -3.0,]})

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C'], exclude_chars=['*'])
    ...  ).round(2)
       site     M     A     C
    0     1  0.47  0.44  0.10
    1     2  0.55  0.31  0.14

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C', '*'], exclude_chars=[])
    ...  ).round(2)
       site     M     A     C     *
    0     1  0.44  0.41  0.09  0.06
    1     2  0.48  0.28  0.12  0.12

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C', '*'], exclude_chars=[],
    ...                  missing='site_average')
    ...  ).round(2)
       site     M     A     C     *
    0     1  0.44  0.41  0.09  0.06
    1     2  0.43  0.25  0.11  0.22

    >>> scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                 alphabet=['M', 'A', 'C', '*'], exclude_chars=[],
    ...                 missing='error')
    Traceback (most recent call last):
        ...
    ValueError: missing functional scores for some mutations

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C'], exclude_chars=['*'],
    ...                  returnformat='tidy')
    ...  ).round(2)
      wildtype  site mutant  preference
    0        M     1      C        0.10
    1        A     2      C        0.14
    2        A     2      A        0.31
    3        M     1      A        0.44
    4        M     1      M        0.47
    5        A     2      M        0.55

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C'], exclude_chars=['*'],
    ...                  stringency_param=3)
    ...  ).round(2)
       site     M     A     C
    0     1  0.55  0.45  0.00
    1     2  0.83  0.16  0.01

    >>> (scores_to_prefs(func_scores_df, 'aa_substitutions', 'func_score', 2,
    ...                  alphabet=['M', 'A', 'C', '*'], exclude_chars=[],
    ...                  returnformat='tidy')
    ...  ).round(2)
      wildtype  site mutant  preference
    0        M     1      *        0.06
    1        M     1      C        0.09
    2        A     2      C        0.12
    3        A     2      *        0.12
    4        A     2      A        0.28
    5        M     1      A        0.41
    6        M     1      M        0.44
    7        A     2      M        0.48

    """
    if not isinstance(exclude_chars, (list, tuple)):
        raise ValueError('`exclude_chars` must be list, tuple (can be empty)')
    exclude_chars = list(exclude_chars)

    alphabet = list(alphabet)
    if [a for a in alphabet if a in exclude_chars]:
        raise ValueError(f"character in `exclude_chars` of {exclude_chars} "
                         f" in `alphabet` of {alphabet}. These lists must be "
                         'mutually exclusive')

    if score_col == mutation_col:
        raise ValueError('`score_col` and `mutation_col` must be different')
    for colname, col in [('mutation', mutation_col), ('score', score_col)]:
        if col not in df.columns:
            raise ValueError(f"`df` lacks `{colname}_col` of {col}")
        if col in {'wildtype', 'site', 'mutant'}:
            raise ValueError(f"`{colname}_col` cannot be named {score_col}")

    if len(df[mutation_col]) != df[mutation_col].nunique():
        raise ValueError('duplicated entries in `mutation_col` of `df`')

    # extract wildtype, site, mutant from mutation
    chars_regex = ''.join(re.escape(a) for a in alphabet + exclude_chars)
    df = (df.join(df
                  [mutation_col]
                  .str
                  .extract(rf"^(?P<wildtype>[{chars_regex}])" +
                           r'(?P<site>\-?\d+)' +
                           rf"(?P<mutant>[{chars_regex}])$")
                  .assign(site=lambda x: x['site'].astype(int))
                  )
          [['site', 'wildtype', 'mutant', score_col, mutation_col]]
          )
    if df.isnull().any().any():
        raise ValueError('unparseable mutations given specified alphabet:\n' +
                         ', '.join(df[mutation_col].tolist()))
    if len(df.query('wildtype == mutant')):
        raise ValueError('`df` contains mutations with same wildtype & mutant')

    # remove any excluded characters
    df = (df
          [['site', 'wildtype', 'mutant', score_col]]
          .query('wildtype not in @exclude_chars')
          .query('mutant not in @exclude_chars')
          )

    # get missing values for later use
    if missing == 'average':
        missing_val = df[score_col].mean()
    elif missing == 'site_average':
        missing_val = df.groupby('site')[score_col].mean().to_dict()
    elif missing != 'error':
        raise ValueError(f"invalid `missing` of {missing}")

    # add wildtype to data frame
    wt_df = (df
             [['wildtype', 'site']]
             .drop_duplicates()
             .assign(mutant=lambda x: x['wildtype'])
             .assign(**{score_col: wt_score})
             )
    df = pd.concat([df, wt_df], sort=False)

    # add missing characters from alphabet for each site as here:
    # https://stackoverflow.com/a/47118819
    mux = pd.MultiIndex.from_product([df['site'].unique(), alphabet],
                                     names=['site', 'mutant'])
    df = (mux
          .to_frame(index=False)
          .assign(wildtype=lambda x: x['site'].map(df
                                                   .set_index('site')
                                                   ['wildtype']
                                                   .to_dict()))
          .merge(df, on=['site', 'wildtype', 'mutant'], how='outer')
          )

    # fill missing values
    if missing == 'average':
        df = df.fillna(missing_val)
    elif missing == 'site_average':
        df_notmissing = df[df.notnull().all(axis=1)]
        df_missing = df[df.isnull().any(axis=1)]
        assert len(df) == len(df_notmissing) + len(df_missing)
        df_missing = (df_missing
                      .assign(**{score_col: lambda x: (x['site']
                                                       .map(missing_val))}
                              )
                      )
        df = pd.concat([df_notmissing, df_missing], sort=False)
    elif df.isnull().any().any():
        raise ValueError('missing functional scores for some mutations')

    # convert to prefs
    df = (df
          .assign(unscaled_prefs=lambda x: (base**x[score_col]
                                            )**stringency_param,
                  preference=lambda x: (x['unscaled_prefs'] /
                                        (x.groupby('site', sort=False)
                                         ['unscaled_prefs']
                                         .transform('sum')
                                         )
                                        )
                  )
          )

    # pivot to wide form
    if returnformat == 'wide':
        df = df.pivot_table(index='site',
                            columns='mutant',
                            values='preference')
        df = df[alphabet]
        df.columns.name = None
        df = df.reset_index()
    elif returnformat == 'tidy':
        df = (df
              [['wildtype', 'site', 'mutant', 'preference']]
              .sort_values('preference')
              .reset_index(drop=True)
              )
    else:
        raise ValueError(f"invalid `returnformat` {returnformat}")
    assert not df.isnull().any().any(), df

    return df


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""Setup script for ``dms_variants``."""


import re
import sys

from setuptools import Extension, setup


if not (sys.version_info[0] == 3 and sys.version_info[1] >= 6):
    raise RuntimeError(
                'dms_variants requires Python >=3.6.\n'
                f"You are using {sys.version_info[0]}.{sys.version_info[1]}.")

# get metadata from package `__init__.py` file as here:
# https://packaging.python.org/guides/single-sourcing-package-version/
metadata = {}
init_file = 'dms_variants/__init__.py'
with open(init_file) as f:
    init_text = f.read()
for dataname in ['version', 'author', 'email', 'url']:
    matches = re.findall(
            '__' + dataname + r'__\s+=\s+[\'"]([^\'"]+)[\'"]',
            init_text)
    if len(matches) != 1:
        raise ValueError(f"found {len(matches)} matches for {dataname} "
                         f"in {init_file}")
    else:
        metadata[dataname] = matches[0]

with open('README.rst') as f:
    readme = f.read()

# main setup command
setup(
    name='dms_variants',
    version=metadata['version'],
    author=metadata['author'],
    author_email=metadata['email'],
    url=metadata['url'],
    download_url='https://github.com/jbloomlab/dms_variants/tarball/' +
                 metadata['version'],  # tagged version on GitHub
    description='Analyze deep mutational scanning of barcoded variants.',
    long_description=readme,
    license='GPLv3',
    install_requires=[
            'biopython>=1.73',
            'matplotlib>=3.1',
            'pandas>=0.25.1',
            'plotnine==0.6',  # https://github.com/has2k1/plotnine/issues/403
            'regex>=2.4.153',
            'scipy>=1.1.0',
            ],
    platforms='Linux and Mac OS X.',
    packages=['dms_variants'],
    package_dir={'dms_variants': 'dms_variants'},
    ext_modules=[
        Extension(
            'dms_variants._cutils',
            ['dms_variants/_cutils.c'],
            extra_compile_args=["-Wno-error=declaration-after-statement"]
            ),
        ],
    )
